import importlib
import importlib.util
from pathlib import Path
import sys
from types import ModuleType
from typing import Dict, Any

DRIVER_SCRIPT = '/python_cwl_driver.py'
TYPES_SCRIPT = '/workflow_types.py'

TYPES_SCRIPT_REL = '../workflow-inference-compiler/examples/scripts/workflow_types.py'

# NOTE: VERY IMPORTANT: Since we have to programmatically import the python file in the compiler,
# and since the act of importing it executes the entire file (i.e. including import statements),

# USERS SHOULD NOT USE TOP-LEVEL IMPORT STATEMENTS!

# See the following links for a more detailed explanation
# https://stackoverflow.com/questions/2724260/why-does-pythons-import-require-fromlist
# https://stackoverflow.com/questions/8790003/dynamically-import-a-method-in-a-file-from-a-string


def import_python_file(python_module_name: str, python_file_path: Path) -> ModuleType:
    """This function import a python file directly, as per the documentation\n
    https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly

    Args:
        python_module_name (str): The name of the python module
        python_file_path (Path): The path to the python file.

    Returns:
        ModuleType: The module that was loaded.
    """
    # NOTE: Apparently import_module resolves symlinks before attempting to import.
    # Since python uses relative paths to determine modules, as you can imagine
    # this causes massive problems. By default, CWL symlinks every single input
    # file into its own temporary directory. If you use an initial workdir, the
    # symlinks are now all in the same directory, but their sources are still
    # pointing wherever. Thus, import_module will never work!
    # module = importlib.import_module(python_script_mod)
    # ModuleNotFoundError: No module named ...

    # The solution is buried in the examples at the bottom of the documentation
    # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    # and https://stackoverflow.com/questions/65206129/importlib-not-utilising-recognising-path
    spec = importlib.util.spec_from_file_location(
        name=python_module_name,  # module name (not file name)
        location=str(python_file_path.absolute())  # ABSOLUTE path!
    )
    if spec:
        module_ = importlib.util.module_from_spec(spec)
        sys.modules[python_module_name] = module_

        try:
            if spec.loader:
                spec.loader.exec_module(module_)  # guard behind if to satisfy mypy
            else:
                raise Exception
        except Exception as e:
            raise Exception(f'Error! Cannot load python_script {python_file_path}') from e
        # Note that now (after calling exec_module) we can call import_module without error
        # module_ = importlib.import_module(python_module_name)
    else:
        raise Exception(f'Error! Cannot load python_script spec {spec} from file\n{python_file_path}')
    return module_


def get_main_args(module_: ModuleType) -> Dict[str, Any]:
    """Uses inspect to get the arguments to the main() function of the given module.

    Args:
        module_ (ModuleType): A ModuleType object returned from import_python_file

    Returns:
        Dict[str, Any]: A dictionary of keys value pairs
    """
    # importing at the top-level causes a circular import error
    # (jsonschema transitively imports inspect)
    import inspect  # pylint: disable=import-outside-toplevel

    anns = inspect.getfullargspec(module_.main).annotations
    ret = {'return': anns.get('return')}  # Separate out the return type
    if 'return' in anns:
        del anns['return']
    # print(anns)
    # print(ret)

    # print(inspect.signature(module_.main).parameters)
    # print(inspect.signature(module_.main).return_annotation)
    return anns


def check_args_match_inputs(module_: ModuleType, args: Dict[str, Any], check: bool = False) -> None:
    """Checks that the keys (only) of the args dict match the keys of the top-level inputs attribute.

    Args:
        module_ (ModuleType): A ModuleType object returned from import_python_file
        args (Dict[str, Any]): A dictionary of keys value pairs
    """
    error = False
    for arg in args:
        if arg not in module_.inputs:
            print(f'Error! wic argument {arg} not in python arguments {module_.inputs}')
            error = True
    # Wait until after inference
    if check:
        for arg in module_.inputs:
            if arg not in args:
                print(f'Error! Python argument {arg} not in wic arguments {args}')
                error = True
    if error:
        sys.exit(1)


def generate_CWL_CommandLineTool(module_inputs: Dict[str, Any], module_outputs: Dict[str, Any],
                                 python_script_docker_pull: str = '') -> Dict[str, Any]:
    """Generates a CWL CommandLineTool for an arbitrary (annotated) python script.

    Args:
        module_inputs (Dict[str, Any]): The top-level inputs attribute of the python module.
        module_outputs (Dict[str, Any]): The top-level inputs attribute of the python module.
        python_script_docker_pull (str): The username/image to use with docker pull ...

    Returns:
        Dict[str, Any]: A CWL CommandLineTool with the given inputs and outputs.
    """
    yaml_tree: Dict[str, Any] = {}
    yaml_tree['cwlVersion'] = 'v1.0'
    yaml_tree['class'] = 'CommandLineTool'
    yaml_tree['$namespaces'] = {'edam': 'https://edamontology.org/'}
    yaml_tree['$schemas'] = ['https://raw.githubusercontent.com/edamontology/edamontology/master/EDAM_dev.owl']
    yaml_tree['baseCommand'] = 'python3'

    types_entry = '$(inputs.workflow_types)'
    driver_entry = '$(inputs.driver_script)'
    script_entry = '$(inputs.script)'
    requirements: Dict[str, Any] = {}
    requirements = {  # 'InitialWorkDirRequirement': {'listing': [script_entry]}, #[types_entry,driver_entry,script_entry]
        'InlineJavascriptRequirement': {}}
    if python_script_docker_pull:
        requirements['DockerRequirement'] = {'dockerPull': python_script_docker_pull}
    yaml_tree['requirements'] = requirements

    def input_binding(position: int, prefix: str = '') -> Dict[str, Any]:
        if prefix == '':
            return {'inputBinding': {'position': position}}
        return {'inputBinding': {'position': position, 'prefix': f'--{prefix}'}}

    inputs: Dict[str, Any] = {}
    # driver_script_file = {'class': 'File', 'path': driver_script}
    inputs['driver_script'] = {'type': 'string', 'format': 'edam:format_2330',
                               **input_binding(1), 'default': DRIVER_SCRIPT}  # driver_script_file
    # workflow_types_file = {'class': 'File', 'path': types_script}
    inputs['workflow_types'] = {'type': 'string', 'format': 'edam:format_2330',
                                **input_binding(2), 'default': TYPES_SCRIPT}  # workflow_types_file
    inputs['script'] = {'type': 'File', 'format': 'edam:format_2330', **input_binding(3)}
    for i, (arg_key, arg_val) in enumerate(module_inputs.items()):
        inputs[arg_key] = {**arg_val, **input_binding(i+4, arg_key)}
    # inputs['args'] = {'type': 'string', **input_binding(4)}
    yaml_tree['inputs'] = inputs

    outputs: Dict[str, Any] = {}
    for i, (arg_key, (glob_pattern, arg_val)) in enumerate(module_outputs.items()):
        outputs[arg_key] = {**arg_val, 'outputBinding': {'glob': glob_pattern}}
    # output_all is optional, but good for debugging bad glob patterns
    output_all = {'type':
                  {'type': 'array',
                   'items': ['Directory', 'File']},
                  'outputBinding': {'glob': '.'},
                  'format': 'edam:format_2330'}  # 'Textual format'
    # This crashes toil-cwl-runner, but not cwltool.
    # outputs['output_all'] = output_all
    yaml_tree['outputs'] = outputs

    yaml_tree['stdout'] = 'stdout'
    return yaml_tree


def get_module(python_script_mod: str, python_script_path: Path, yml_args: Dict[str, Any]) -> ModuleType:
    """Imports the given python script and validates its top-level annotations.

    Args:
        python_script_mod (str): The module name of the given python script.
        python_script_path (Path): The path to the given python script.
        yml_args (Dict[str, Any]): The contents of the python_script in: yml tag.

    Returns:
        ModuleType: The Module object associated with the given python script.
    """
    import_python_file('workflow_types', Path(TYPES_SCRIPT_REL))
    module_ = import_python_file(python_script_mod, python_script_path)
    # print(module_.inputs)
    # print(module_.outputs)

    main_args = get_main_args(module_)
    check_args_match_inputs(module_, main_args)
    # TODO: validate module_.inputs values (check types and formats, and nothing else)
    # TODO: validate module_.outputs

    check_args_match_inputs(module_, yml_args)
    return module_


def get_inputs_workflow(module_inputs: Dict[str, Any], python_script_path: str,
                        yml_args: Dict[str, Any]) -> Dict[str, Any]:
    """This generates the contents of the inputs file associated with generate_CWL_CommandLineTool\n
    Note that this is already taken care of in the compiler, but this function\n
    is useful for standalone purposes. (Alternatively, just make a single-step workflow.)

    Args:
        module_inputs (Dict[str, Any]): The top-level inputs attribute of the python module.
        python_script_path (str): The path to the given python script.
        yml_args (Dict[str, Any]): The contents of the python_script in: yml tag.

    Returns:
        Dict[str, Any]: The contents of the CWL inputs file.
    """
    inputs_workflow = {}
    inputs_workflow['script'] = {'class': 'File', 'format': 'edam:format_2330', 'path': python_script_path}
    for i, (arg, yml_val) in enumerate(yml_args.items()):
        if module_inputs[arg]['type'] == 'string':
            inputs_workflow[arg] = yml_val
        else:
            inputs_workflow[arg] = {'class': 'File', 'format': module_inputs[arg]['format'], 'path': yml_val}
    # inputs_workflow = {'script': f'{python_script}.py', **yml_args}
    # inputs_workflow = {'script': f'{python_script}.py', 'args': json.dumps(yml_args)}
    return inputs_workflow



import argparse
import glob
import json
import subprocess as sub
import sys
import os
import stat
from pathlib import Path
import shutil
import platform
import traceback
from typing import Dict, List, Optional

try:
    import cwltool.main
    import toil.cwl.cwltoil  # transitively imports cwltool
except ImportError as exc:
    print('Could not import cwltool.main and/or toil.cwl.cwltoil')
    # (pwd is imported transitively in cwltool.provenance)
    print(exc)
    if exc.msg == "No module named 'pwd'":
        print('Windows does not have a pwd module')
        print('If you want to run on windows, you need to install')
        print('Windows Subsystem for Linux')
        print('See https://pypi.org/project/cwltool/#ms-windows-users')
    else:
        raise exc

from . import auto_gen_header
from . import utils  # , utils_graphs
from .wic_types import Yaml, RoseTree
from .plugins import logging_filters


def generate_run_script(cmdline: str) -> None:
    """Writes the command used to invoke the cwl-runner to run.sh
    Does not actually invoke ./run.sh

    Args:
        cmdline (str): The command line which invokes the cwl-runner
    """
    with open('run.sh', mode='w', encoding='utf-8') as f:
        f.write('#!/bin/bash -e\n')
        f.write(auto_gen_header)
        f.write(cmdline)
    # chmod +x run.sh
    st = os.stat('run.sh')
    os.chmod('run.sh', st.st_mode | stat.S_IEXEC)


def run_local(args: argparse.Namespace, rose_tree: RoseTree, cachedir: Optional[str], cwl_runner: str,
              use_subprocess: bool) -> int:
    """This function runs the compiled workflow locally.

    Args:
        args (argparse.Namespace): The command line arguments
        rose_tree (RoseTree): The compiled workflow
        cachedir (Optional[str]): The --cachedir to use (if any)
        cwl_runner (str): Either 'cwltool' or 'toil-cwl-runner'
        use_subprocess (bool): When using cwltool, determines whether to use subprocess.run(...)
        or use the cwltool python api.

    Returns:
        retval: The return value
    """

    docker_cmd: str = args.container_engine
    # Check that docker is installed, so users don't get a nasty runtime error.
    cmd = [docker_cmd, 'run', '--rm', 'hello-world']
    output = ''
    try:
        docker_cmd_exists = True
        proc = sub.run(cmd, check=False, stdout=sub.PIPE, stderr=sub.STDOUT)
        output = proc.stdout.decode("utf-8")
    except FileNotFoundError:
        docker_cmd_exists = False
    out_d = "Hello from Docker!"
    out_p = "Hello Podman World"
    permission_denied = 'permission denied while trying to connect to the Docker daemon socket at'
    if (not docker_cmd_exists or not (proc.returncode == 0 and out_d in output or out_p in output)) and not args.ignore_docker_install:
        if permission_denied in output:
            print('Warning! docker appears to be installed, but not configured as a non-root user.')
            print('See https://docs.docker.com/engine/install/linux-postinstall/#manage-docker-as-a-non-root-user')
            print('TL;DR you probably just need to run the following command (and then restart your machine)')
            print('sudo usermod -aG docker $USER')
            sys.exit(1)
        print(f'Warning! The {docker_cmd} command does not appear to be installed.')
        print(f'Most workflows require docker containers and will fail at runtime if {docker_cmd} is not installed.')
        print('If you want to try running the workflow anyway, use --ignore_docker_install')
        print('Note that --ignore_docker_install does NOT change whether or not any step in your workflow uses docker')
        sys.exit(1)

    # If docker is installed, check for too many running processes. (on linux, macos)
    if docker_cmd == 'docker' and docker_cmd_exists and sys.platform != "win32":
        cmd = 'pgrep com.docker | wc -l'  # type: ignore
        proc = sub.run(cmd, check=False, stdout=sub.PIPE, stderr=sub.STDOUT, shell=True)
        output = proc.stdout.decode("utf-8")
        num_processes = int(output.strip())
        max_processes = 1000
        if num_processes > max_processes and not args.ignore_docker_processes:
            print(f'Warning! There are {num_processes} running docker processes.')
            print(f'More than {max_processes} may potentially cause intermittent hanging issues.')
            print('It is recommended to terminate the processes using the command')
            print('`sudo pkill com.docker && sudo pkill Docker`')
            print('and then restart Docker.')
            print('If you want to run the workflow anyway, use --ignore_docker_processes')
            sys.exit(1)

    yaml_path = args.yaml
    yaml_stem = Path(args.yaml).stem

    yaml_inputs = rose_tree.data.workflow_inputs_file
    stage_input_files(yaml_inputs, Path(args.yaml).parent.absolute())

    retval = 1  # overwrite if successful
    provenance: List[str] = []
    # NOTE: By default, cwltool will attempt to download schema files.
    # $schemas:
    #   - https://raw.githubusercontent.com/edamontology/edamontology/master/EDAM_dev.owl
    # If you have connection issues (e.g. firewall, VPN, etc) then failure to download will
    # not actually cause any problems immediately (other than a ~30 second timeout).
    # However, cwltool does not appear to cache these files, so it will attempt to download
    # them repeatedly. These ~30 second timeouts will eventually add up to >6 hours, which
    # will cause github to terminate the CI Action...
    skip_schemas = ['--skip-schemas'] if not args.no_skip_dollar_schemas else []

    yaml_stem = yaml_stem + '_inline' if args.cwl_inline_subworkflows else yaml_stem
    if cwl_runner == 'cwltool':
        parallel = ['--parallel'] if args.parallel else []
        # NOTE: --parallel is required for real-time analysis / real-time plots,
        # but it seems to cause hanging with Docker for Mac. The hanging seems
        # to be worse when using parallel scattering.
        quiet = ['--quiet'] if args.quiet else []
        cachedir_ = ['--cachedir', cachedir] if cachedir else []
        net = ['--custom-net', args.custom_net] if args.custom_net else []
        provenance = ['--provenance', f'provenance/{yaml_stem}'] if not args.no_provenance else []
        docker_cmd_ = [] if docker_cmd == 'docker' else ['--user-space-docker-cmd', docker_cmd]
        write_summary = ['--write-summary', f'output_{yaml_stem}.json']
        path_check = ['--relax-path-checks']
        # See https://github.com/common-workflow-language/cwltool/blob/5a645dfd4b00e0a704b928cc0bae135b0591cc1a/cwltool/command_line_tool.py#L94
        # NOTE: Using --leave-outputs to disable --outdir
        # See https://github.com/dnanexus/dx-cwl/issues/20
        # --outdir has one or more bugs which will cause workflows to fail!!!
        docker_pull = ['--disable-pull']  # Use cwl-docker-extract to pull images
        script = 'cwltool_filterlog_pf' if args.partial_failure_enable else 'cwltool_filterlog'
        cmd = [script] + docker_pull + parallel + quiet + cachedir_ + net + provenance + \
            docker_cmd_ + write_summary + skip_schemas + path_check
        cmd += ['--leave-outputs',
                # '--js-console', # "Running with support for javascript console in expressions
                # (DO NOT USE IN PRODUCTION)"
                f'autogenerated/{yaml_stem}.cwl', f'autogenerated/{yaml_stem}_inputs.yml']
        # TODO: Consider using the undocumented flag --fast-parser for known-good workflows,
        # which was recently added in the 3.1.20220913185150 release of cwltool.
        cmdline = ' '.join(cmd)

        if args.generate_run_script:
            generate_run_script(cmdline)
            return 0  # Do not actually run

        # If we are actually running, then enable --copy_output_files
        args.copy_output_files = True

        print('Running ' + cmdline)
        if use_subprocess:
            # To run in parallel (i.e. pytest ... --workers 8 ...), we need to
            # use separate processes. Otherwise:
            # "signal only works in main thread or with __pypy__.thread.enable_signals()"
            proc = sub.run(cmd, check=False)
            retval = proc.returncode
            return retval  # Skip copying files to outdir/ for CI
        else:
            print('via cwltool.main.main python API')
            try:
                retval = cwltool.main.main(cmd[1:])

                print(f'Final output json metadata blob is in output_{yaml_stem}.json')

                # See https://pypi.org/project/cwltool/#import-as-a-module
                # This also works, but doesn't easily allow using --leave-outputs, --provenence, --cachedir
                # import cwltool.factory
                # fac = cwltool.factory.Factory()
                # rootworkflow = fac.make(f'autogenerated/{yaml_stem}.cwl')
                # output_json = rootworkflow(**yaml_inputs)
                # with open('primary-output.json', mode='w', encoding='utf-8') as f:
                #     f.write(json.dumps(output_json))
            except Exception as e:
                print('Failed to execute', yaml_path)
                print(f'See error_{yaml_stem}.txt for detailed technical information.')
                # Do not display a nasty stack trace to the user; hide it in a file.
                with open(f'error_{yaml_stem}.txt', mode='w', encoding='utf-8') as f:
                    # https://mypy.readthedocs.io/en/stable/common_issues.html#python-version-and-system-platform-checks
                    if sys.version_info >= (3, 10):
                        traceback.print_exception(type(e), value=e, tb=None, file=f)
                    else:
                        traceback.print_exception(etype=type(e), value=e, tb=None, file=f)
                if not cachedir:  # if running on CI
                    print(e)

    if cwl_runner == 'toil-cwl-runner':
        if platform.python_implementation().lower() == 'pypy':
            print('Error! Toil is not compatible with pypy!')
            print('Please use the standard cpython interpreter with Toil.')
            sys.exit(1)

        # NOTE: toil-cwl-runner always runs in parallel
        net = ['--custom-net', args.custom_net] if args.custom_net else []
        provenance = ['--provenance', f'provenance/{yaml_stem}'] if not args.no_provenance else []
        docker_cmd_ = [] if docker_cmd == 'docker' else ['--user-space-docker-cmd', docker_cmd]
        path_check = ['--relax-path-checks']
        # See https://github.com/common-workflow-language/cwltool/blob/5a645dfd4b00e0a704b928cc0bae135b0591cc1a/cwltool/command_line_tool.py#L94
        # https://github.com/DataBiosphere/toil/blob/6558c7f97fb37c6ef6f469c7ae614109050322f4/src/toil/options/cwl.py#L152
        docker_pull = []  # toil supports --force-docker-pull, but not --disable-pull
        cmd = ['toil-cwl-runner'] + docker_pull + net + provenance + docker_cmd_ + path_check
        cmd += ['--outdir', 'outdir_toil',
                '--jobStore', f'file:./jobStore_{yaml_stem}',  # NOTE: This is the equivalent of --cachedir
                # TODO: Check --clean, --cleanWorkDir, --restart
                '--clean', 'always',  # This effectively disables caching, but is reproducible
                f'autogenerated/{yaml_stem}.cwl', f'autogenerated/{yaml_stem}_inputs.yml']
        cmdline = ' '.join(cmd)

        if args.generate_run_script:
            generate_run_script(cmdline)
            return 0  # Do not actually run

        print('Running ' + cmdline)
        if use_subprocess:
            # To run in parallel (i.e. pytest ... --workers 8 ...), we need to
            # use separate processes. Otherwise:
            # "signal only works in main thread or with __pypy__.thread.enable_signals()"
            proc = sub.run(cmd, check=False)
            retval = proc.returncode
            return retval  # Skip copying files to outdir/ for CI
        else:
            print('via toil.cwl.cwltoil.main python API')
            try:
                retval = toil.cwl.cwltoil.main(cmd[1:])

            except Exception as e:
                retval = 1
                print('Failed to execute', yaml_path)
                print(f'See error_{yaml_stem}.txt for detailed technical information.')
                # Do not display a nasty stack trace to the user; hide it in a file.
                with open(f'error_{yaml_stem}.txt', mode='w', encoding='utf-8') as f:
                    if sys.version_info >= (3, 10):
                        traceback.print_exception(type(e), value=e, tb=None, file=f)
                    else:
                        traceback.print_exception(etype=type(e), value=e, tb=None, file=f)
                if not cachedir:  # if running on CI
                    print(e)

    if retval == 0:
        print('Success! Output files should be in outdir/')
    else:
        print('Failure! Please scroll up and find the FIRST error message.')
        print('(You may have to scroll up A LOT.)')

    # Remove the annoying cachedir* directories that somehow aren't getting automatically deleted.
    # NOTE: Do NOT allow cachedir to be absolute; otherwise
    # if users pass in "/" this will delete their entire hard drive.
    cachedir_path = str(cachedir)
    if not Path(cachedir_path).is_absolute():
        for d in glob.glob(cachedir_path + '*'):
            if not d == cachedir_path:
                shutil.rmtree(d)  # Be VERY careful when programmatically deleting directories!

    return retval


def copy_output_files(yaml_stem: str) -> None:
    """Copies output files from the cachedir to outdir/

    Args:
        yaml_stem (str): The --yaml filename (without .extension)
    """
    output_json_file = Path(f'output_{yaml_stem}.json')
    if output_json_file.exists():
        pass  # TODO

    output_json_file_prov = Path(f'provenance/{yaml_stem}/workflow/primary-output.json')
    # NOTE: The contents of --write-summary is
    # slightly different than provenance/{yaml_stem}/workflow/primary-output.json!
    # They are NOT the same file!
    if output_json_file_prov.exists():
        with open(output_json_file_prov, mode='r', encoding='utf-8') as f:
            output_json = json.loads(f.read())
        files = utils.parse_provenance_output_files(output_json)

        dests = set()
        for location, namespaced_output_name, basename in files:
            try:
                yaml_stem_init, shortened = utils.shorten_namespaced_output_name(namespaced_output_name)
                parentdirs = yaml_stem_init + '/' + shortened.replace('___', '/')
            except:
                parentdirs = namespaced_output_name  # For --allow_raw_cwl
            Path('outdir/' + parentdirs).mkdir(parents=True, exist_ok=True)
            source = f'provenance/{yaml_stem}/workflow/' + location
            # NOTE: Even though we are using subdirectories (not just a single output directory),
            # there is still the possibility of filename collisions, i.e. when scattering.
            # For now, let's use a similar trick as cwltool of append _2, _3 etc.
            # except do it BEFORE the extension.
            # This could still cause problems with slicing, i.e. if you scatter across
            # indices 11-20 first, then 1-10 second, the output file indices will get switched.
            dest = 'outdir/' + parentdirs + '/' + basename
            if dest in dests:
                idx = 2
                while Path(dest).exists():
                    stem = Path(basename).stem
                    suffix = Path(basename).suffix
                    dest = 'outdir/' + parentdirs + '/' + stem + f'_{idx}' + suffix
                    idx += 1
            dests.add(dest)
            cmd = ['cp', source, dest]
            sub.run(cmd, check=True)


def stage_input_files(yml_inputs: Yaml, root_yml_dir_abs: Path,
                      relative_run_path: bool = True, throw: bool = True) -> None:
    """Copies the input files in yml_inputs to the working directory.

    Args:
        yml_inputs (Yaml): The yml inputs file for the root workflow.
        root_yml_dir_abs (Path): The absolute path of the root workflow yml file.
        relative_run_path (bool): Controls whether to use subdirectories or\n
        just one directory when writing the compiled CWL files to disk
        throw (bool): Controls whether to raise/throw a FileNotFoundError.

    Raises:
        FileNotFoundError: If throw and it any of the input files do not exist.
    """
    for key, val in yml_inputs.items():
        if isinstance(val, Dict) and val.get('class', '') == 'File':
            path = root_yml_dir_abs / Path(val['path'])
            if not path.exists() and throw:
                # raise FileNotFoundError(f'Error! {path} does not exist!')
                print(f'Error! {path} does not exist!')
                sys.exit(1)

            relpath = Path('autogenerated/') if relative_run_path else Path('.')
            pathauto = relpath / Path(val['path'])  # .name # NOTE: Use .name ?
            pathauto.parent.mkdir(parents=True, exist_ok=True)

            if path != pathauto:
                cmd = ['cp', str(path), str(pathauto)]
                proc = sub.run(cmd, check=False)


def cwltool_main() -> int:
    """Another entrypoint for filtering regular logged output"""
    logging_filters()
    cmd = sys.argv
    retval: int = cwltool.main.main(cmd[1:])
    return retval


def cwltool_main_pf() -> int:
    """Another entrypoint for filtering partial failures logged output"""
    logging_filters(True)
    cmd = sys.argv
    retval: int = cwltool.main.main(cmd[1:])
    return retval



import copy
from typing import Any, Dict, List

from jsonschema import Draft202012Validator

# from sophios import cli
from sophios.wic_types import Json

SCHEMA: Json = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "definitions": {
        "Link": {
            "properties": {
                "id": {
                    "type": "number"
                },
                "inletIndex": {
                    "type": "number"
                },
                "outletIndex": {
                    "type": "number"
                },
                "sourceId": {
                    "type": "number"
                },
                "targetId": {
                    "type": "number"
                },
                "x1": {
                    "type": "number"
                },
                "x2": {
                    "type": "number"
                },
                "y1": {
                    "type": "number"
                },
                "y2": {
                    "type": "number"
                }
            },
            "type": "object",
            "required": ["id", "sourceId", "targetId"]
        },
        "NodeSettings": {
            "properties": {
                "inputs": {
                    "additionalProperties": {
                        "$ref": "#/definitions/T"
                    },
                    "type": "object"
                },
                "outputs": {
                    "additionalProperties": {
                        "$ref": "#/definitions/T"
                    },
                    "type": "object"
                }
            },
            "type": "object"
        },
        "NodeX": {
            "properties": {
                "expanded": {
                    "type": "boolean"
                },
                "height": {
                    "type": "number"
                },
                "id": {
                    "type": "number"
                },
                "internal": {
                    "type": "boolean"
                },
                "name": {
                    "type": "string"
                },
                "pluginId": {
                    "type": "string"
                },
                "settings": {
                    "$ref": "#/definitions/NodeSettings"
                },
                "width": {
                    "type": "number"
                },
                "x": {
                    "type": "number"
                },
                "y": {
                    "type": "number"
                },
                "z": {
                    "type": "number"
                },
            },
            "type": "object",
            "required": ["id", "name", "pluginId", "settings", "internal"]
        },
        "T": {
            "type": "object"
        }
    },
    "properties": {
        "links": {
            "items": {
                "$ref": "#/definitions/Link"
            },
            "type": "array"
        },
        "nodes": {
            "items": {
                "$ref": "#/definitions/NodeX"
            },
            "type": "array"
        },
        "selection": {
            "items": {
                "type": "number"
            },
            "type": "array"
        }
    },
    "type": "object",
    "required": ["links", "nodes"]
}


def del_irrelevant_keys(ldict: List[Dict[Any, Any]], relevant_keys: List[Any]) -> None:
    """deletes irrelevant keys from every dict in the list of dicts"""
    for elem in ldict:
        ekeys = list(elem.keys())
        for ek in ekeys:
            if ek not in relevant_keys:
                # delete the key if it exists
                elem.pop(ek, None)


def validate_schema_and_object(schema: Json, jobj: Json) -> None:
    """Validate schema object"""
    Draft202012Validator.check_schema(schema)
    df2012 = Draft202012Validator(schema)
    df2012.is_valid(jobj)


def raw_wfb_to_lean_wfb(inp: Json) -> Json:
    """drop all the unnecessary info from incoming wfb object"""
    inp_restrict = copy.deepcopy(inp)
    keys = list(inp.keys())
    # To avoid deserialization
    # required attributes from schema
    prop_req = SCHEMA['required']
    nodes_req = SCHEMA['definitions']['NodeX']['required']
    links_req = SCHEMA['definitions']['Link']['required']
    do_not_rem_nodes_prop = ['cwlScript']
    do_not_rem_links_prop: list = []

    for k in keys:
        if k not in prop_req:
            del inp_restrict[k]
        elif k == 'links':
            lems = inp_restrict[k]
            rel_links_keys = links_req + do_not_rem_links_prop
            del_irrelevant_keys(lems, rel_links_keys)
        elif k == 'nodes':
            nems = inp_restrict[k]
            rel_nodes_keys = nodes_req + do_not_rem_nodes_prop
            del_irrelevant_keys(nems, rel_nodes_keys)
        else:
            pass

    validate_schema_and_object(SCHEMA, inp_restrict)
    return inp_restrict


# def wfb_to_wic(request: Json) -> Json:
#     """Convert the json object from http request object to a json object that can be used as input to wic compiler.

#     Args:
#         request (Json): json object from http request

#     Returns:
#         converted_json (Json): json object that can be used as input to wic compiler"""

#     converted_steps: list[Any] = []

#     for step in request['steps']:
#         step_template = step['template']
#         arguments = step['arguments']
#         # Get the template name from the step template
#         template_name = next((tval['name']
#                              for tname, tval in request['templates'].items() if step_template == tname), None)
#         # template_name = None
#         # for tname, tval in request['templates'].items():
#         #     if tname == step_template and tval['name']:
#         #         template_name = tval['name']
#         #         break
#         #     elif tname == step_template and not tval['name']:
#         #         break
#         #     else:
#         #         pass

#         converted_step: Json = {}
#         if template_name:
#             converted_step[template_name] = {
#                 "in": {}
#             }

#         for key, value in arguments.items():
#             # be aware of usage of periods in the values as delimiters, this may cause an issue when storing in MongoDB
#             if value.startswith("steps."):
#                 parts = value.split('.')
#                 src_step_idx = int(parts[1][len("step"):])
#                 src_output_key = parts[3]

#                 # Get the src template name from the src step name
#                 src_template = next((step.get("template")
#                                     for step in request['steps'] if step.get("name") == parts[1]), None)
#                 src_template_name = next((stval['name'] for stname, stval in request['templates'].items()
#                                           if src_template == stname), None)
#                 src_converted_step = next((step.get(src_template_name)
#                                           for step in converted_steps if step.get(src_template_name)), None)
#                 if src_converted_step:
#                     src_converted_step["in"][src_output_key] = f"&{src_template_name}.{src_output_key}.{src_step_idx}"
#                     converted_step[template_name]["in"][key] = f"*{src_template_name}.{src_output_key}.{src_step_idx}"
#             else:
#                 converted_step[template_name]["in"][key] = value
#         converted_steps.append(converted_step)

#     converted_json: Json = {"steps": converted_steps}
#     return converted_json



import copy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import yaml

from .wic_types import (Namespaces, RoseTree, StepId, Json, Yaml, YamlForest, YamlTree)


def step_name_str(yaml_stem: str, i: int, step_key: str) -> str:
    """Returns a string which uniquely and hierarchically identifies a step in a workflow

    Args:
        yaml_stem (str): The name of the workflow (filepath stem)
        i (int): The (zero-based) step number
        step_key (str): The name of the step (used as a dict key)

    Returns:
        str: The parameters (and the word 'step') joined together with double underscores
    """
    # Use double underscore so we can '__'.split() below.
    # (This should work as long as yaml_stem and step_key do not contain __)
    return f'{yaml_stem}__step__{i+1}__{step_key}'


def parse_step_name_str(step_name: str) -> Tuple[str, int, str]:
    """The inverse function to step_name_str()

    Args:
        step_name (str): A string of the same form as returned by step_name_str()

    Raises:
        Exception: If the argument is not of the same form as returned by step_name_str()

    Returns:
        Tuple[str, int, str]: The parameters used to create step_name
    """
    vals = step_name.split('__')  # double underscore
    if not len(vals) == 4:
        raise Exception(f"Error! {step_name} is not of the format \n"
                        + '{yaml_stem}__step__{i+1}__{step_key}\n'
                        + 'yaml_stem and step_key should not contain any double underscores.')
    try:
        i = int(vals[2])
    except Exception as ex:
        raise Exception(f"Error! {step_name} is not of the format \n"
                        + '{yaml_stem}__step__{i+1}__{step_key}') from ex
    return (vals[0], i-1, vals[3])


def shorten_namespaced_output_name(namespaced_output_name: str, sep: str = ' ') -> Tuple[str, str]:
    """Removes the intentionally redundant yaml_stem prefixes from the list of
    step_name_str's embedded in namespaced_output_name which allows each
    step_name_str to be context-free and unique. This is potentially dangerous,
    and the only purpose is so we can slightly shorten the output filenames.

    Args:
        namespaced_output_name (str): A string of the form:
        '___'.join(namespaces + [step_name_i, out_key])
        sep (str): The separator used to construct the shortened step name strings.

    Returns:
        Tuple[str, str]: the first yaml_stem, so this function can be inverted,
        and namespaced_output_name, with the embedded yaml_stem prefixes
        removed and double underscores replaced with a single space.
    """
    split = namespaced_output_name.split('___')
    namespaces = split[:-1]
    output_name = split[-1]
    strs = []
    yaml_stem_init = ''
    if len(namespaces) > 0:
        yaml_stem_init = parse_step_name_str(namespaces[0])[0]
        for stepnamestr in namespaces:
            _, i, step_key = parse_step_name_str(stepnamestr)
            strs.append(f'step{sep}{i+1}{sep}{step_key}')
    shortened = '___'.join(strs + [output_name])
    return (yaml_stem_init, shortened)


def restore_namespaced_output_name(yaml_stem_init: str, shortened_output_name: str, sep: Optional[str] = None) -> str:
    """The inverse function to shorten_namespaced_output_name()

    Args:
        yaml_stem_init (str): The initial yaml_stem prefix
        shortened_output_name (str): The shortened namespaced_output_name
        sep (Optional[str], optional): The separator used for shortening. Defaults to None.

    Raises:
        Exception: If the argument is not of the same form as returned by shorten_namespaced_output_name

    Returns:
        str: The original namespaced_output_name before shortening.
    """
    if yaml_stem_init == '':
        return shortened_output_name
    else:
        split = shortened_output_name.split('___')
        namespaces = split[:-1]
        output_name = split[-1]
        yaml_stem = yaml_stem_init
        strs = []
        for shortened_step_name_str in namespaces:
            words = shortened_step_name_str.split(sep)
            if len(words) != 3:
                raise Exception(f'Error! {shortened_step_name_str} is not of the correct format!')
            _, num, name_yml = words
            strs.append(f'{yaml_stem}{sep}step{sep}{num}{sep}{name_yml}')
            yaml_stem = Path(name_yml).stem
        restored = '___'.join(strs + [output_name])
        return restored


def partition_by_lowest_common_ancestor(nss1: Namespaces, nss2: Namespaces) -> Tuple[Namespaces, Namespaces]:
    """See https://en.wikipedia.org/wiki/Lowest_common_ancestor

    Args:
        nss1 (Namespaces): The namespaces associated with the first node
        nss2 (Namespaces): The namespaces associated with the second node

    Returns:
        Tuple[Namespaces, Namespaces]: nss1, partitioned by lowest common ancestor
    """
    # Only partition nss1; if you want to partition nss1
    # just switch the arguments at the call site.
    if nss1 == [] or nss2 == []:
        return ([], nss1)  # Base case
    if nss1[0] == nss2[0]:  # Keep going
        (nss1_heads, nss1_tails) = partition_by_lowest_common_ancestor(nss1[1:], nss2[1:])
        return ([nss1[0]] + nss1_heads, nss1_tails)
    return ([], nss1)


def get_steps_keys(steps: List[Yaml]) -> List[str]:
    """Returns the name (dict key) of each step in the given CWL workflow

    Args:
        steps (List[Yaml]): The steps: tag of a CWL workflow

    Returns:
        List[str]: The name of each step in the given CWL workflow
    """
    steps_keys = [step_dict['id'] for step_dict in steps]
    # print(steps_keys)
    return steps_keys


def get_subkeys(steps_keys: List[str]) -> List[str]:
    """This function determines which step keys are associated with subworkflows.\n
    This is critical for the control flow in many areas of the compiler.

    Args:
        steps_keys (List[str]): All of the step keys for the current workflow.

    Returns:
        List[str]: The list of step keys associated with subworkflows of the current workflow.
    """
    return [key for key in steps_keys if key.endswith('.wic')]


def extract_implementation(yaml_tree: Yaml, wic: Yaml, yaml_path: Path) -> Tuple[str, Yaml]:
    """Chooses a specific implementation for a given CWL workflow step.

    The implementations should be thought of as either 'exactly' identical, or at
    least the same high-level protocol but implemented with a different algorithm.

    Args:
        yaml_tree (Yaml): A Yaml AST dict with sub-dicts for each implementation.
        yaml_path (Path): The filepath of yaml_tree, only used for error reporting.

    Raises:
        Exception: If the steps: and/or implementation: tags are not present.

    Returns:
        Tuple[str, Yaml]: The Yaml AST dict of the chosen implementation.
    """
    yaml_tree_copy = copy.deepcopy(yaml_tree)
    implementation = ''
    if 'implementations' in wic:
        if 'default_implementation' in wic:
            implementation = wic['default_implementation']
        if 'implementation' in wic:
            implementation = wic['implementation']
        if implementation == '':
            raise Exception(f'Error! No implementation in {yaml_path}!')

        plugin_ns = wic.get('namespace', 'global')
        stepid = StepId(implementation, plugin_ns)
        if stepid not in wic['implementations']:
            print(yaml.dump(yaml_tree))
            print(yaml.dump(wic))
            print(wic['implementations'])
            raise Exception(f'Error! No steps for implementation {stepid} in {yaml_path}!')
        steps = wic['implementations'][stepid]['steps']
        yaml_tree_copy.update({'steps': steps})
        # TODO: Use the entire back_tree? Useful for inputs:
        # back_tree = wic['implementations'][stepid]
        # if 'wic' in back_tree:
        #    del back_tree['wic']
        # yaml_tree_copy.update(back_tree)
    elif 'steps' in yaml_tree_copy:
        pass  # steps = yaml_tree_copy['steps']
    else:
        raise Exception(f'Error! No implementations and/or steps in {yaml_path}!')
    return (implementation, yaml_tree_copy)


def flatten(lists: List[List[Any]]) -> List[Any]:
    """Concatenates a list of lists into a single list.

    Args:
        lists (List[List[Any]]): A list of lists

    Returns:
        List[Any]: A single list
    """
    return [x for lst in lists for x in lst]


def flatten_rose_tree(rose_tree: RoseTree) -> List[Any]:
    """Flattens the data contained in the Rose Tree into a List

    Args:
        rose_tree (RoseTree): A Rose Tree

    Returns:
        List[Any]: The list of data associated with each node in the RoseTree
    """
    sub_rose_trees = [flatten_rose_tree(r) for r in rose_tree.sub_trees]
    return [rose_tree.data] + flatten(sub_rose_trees)


def pretty_print_forest(forest: YamlForest) -> None:
    """pretty prints a YamlForest

    Args:
        forest (YamlForest): The forest to be printed
    """
    print(forest.yaml_tree.step_id)
    print(yaml.dump(forest.yaml_tree.yml))
    print(yaml.dump(forest.sub_forests))


def flatten_forest(forest: YamlForest) -> List[YamlForest]:
    """Flattens the sub-trees encountered while traversing an AST

    Args:
        forest (YamlForest): The yaml AST forest to be flattened

    Raises:
        Exception: If implementation: tags are missing.

    Returns:
        List[YamlForest]: The flattened forest
    """
    # pretty_print_forest(forest)
    if forest == {}:
        return []
    yaml_tree = forest.yaml_tree.yml
    wic = {'wic': yaml_tree.get('wic', {})}
    plugin_ns = wic['wic'].get('namespace', 'global')

    if 'implementations' in wic['wic']:
        # pretty_print_forest(forest)
        back_name = ''
        if 'default_implementation' in wic['wic']:
            back_name = wic['wic']['default_implementation']
        if 'implementation' in wic['wic']:
            back_name = wic['wic']['implementation']
        if back_name == '':
            pretty_print_forest(forest)
            raise Exception('Error! No implementation in yaml forest!\n')
        sub_forests_dict = dict(forest.sub_forests)
        step_id = StepId(back_name, plugin_ns)
        yaml_tree_back: YamlTree = sub_forests_dict[step_id].yaml_tree
        step_1 = yaml_tree_back.yml['steps'][0]
        step_name_1 = list(step_1.keys())[0]
        if Path(step_name_1).suffix == '.wic':
            # Choose a specific implementation
            return flatten_forest(sub_forests_dict[step_id])
        return [forest]

    forests = [f[1] for f in forest.sub_forests]
    sub_forests = [flatten_forest(f) for f in forests]
    # Use depth first search flattening to match flatten_rose_tree()
    # bfs = forests + flatten(sub_forests)
    dfs_lists = [[f] + fs for f, fs in zip(forests, sub_forests)]
    dfs = flatten(dfs_lists)
    return dfs


def recursively_delete_dict_key(key: str, obj: Any) -> Any:
    """Recursively deletes any dict entries with the given key.

    Args:
        key (str): The key to be deleted
        obj (Any): The object from which to delete key.

    Returns:
        Any: The original dict with the given key recursively deleted.
    """
    if isinstance(obj, List):
        return [recursively_delete_dict_key(key, x) for x in obj]
    if isinstance(obj, Dict):
        new_dict = {}
        for key_ in obj.keys():
            if not key_ == key:  # i.e. effectively delete key
                new_dict[key_] = recursively_delete_dict_key(key, obj[key_])
        return new_dict
    return obj


def recursively_contains_dict_key(key: str, obj: Any) -> bool:
    """Recursively checks whether obj contains entries with the given key.

    Args:
        key (str): The key to be checked
        obj (Any): The object from which to check the key.

    Returns:
        bool: True if key is found, else False.
    """
    if isinstance(obj, List):
        return any([recursively_contains_dict_key(key, x) for x in obj])
    if isinstance(obj, Dict):
        return (key in obj.keys()) or any(recursively_contains_dict_key(key, val) for val in obj.values())
    return False


def parse_int_string_tuple(string: str) -> Tuple[int, str]:
    """Parses a string of the form '(int, string)'

    Args:
        string (str): A string with the above encoding

    Returns:
        Tuple[int, str]: The parsed result
    """
    string_no_parens = string.strip()[1:-1]
    (str1, str2) = string_no_parens.split(',')
    return (int(str1.strip()), str2.strip())


def reindex_wic_steps(wic_steps: Yaml, index: int, num_steps: int = 1) -> Yaml:
    """ Increment 1-based step index starting from the step with the given index by num_steps.

    This function can be used to reindex steps after inserting num_steps at the given index: in\n
    the wic: metadata annotations tag whose index (before insertion) is >= the given index.

    Args:
        wic_steps (Yaml): The steps: subtag of the wic: metadata annotations tag.
        index (int): The (one-based) start index that needs to be reindexed.
        num_steps (int): The number of steps inserted.

    Returns:
        Yaml: The updated wic: steps: tag, with the appropriate indices incremented.
    """
    wic_steps_reindexed = {}
    for keystr, val in wic_steps.items():
        (i, s) = parse_int_string_tuple(keystr)
        newstr = f'({i+num_steps}, {s})' if i >= index else keystr
        wic_steps_reindexed[newstr] = val
    return wic_steps_reindexed


def get_step_name_1(step_1_names: List[str],
                    yaml_stem: str,
                    namespaces: Namespaces,
                    steps_keys: List[str],
                    subkeys: List[str]) -> str:
    """Finds the name of the first step in the current subworkflow. If the first
    step is itself subworkflow, the call site recurses until it finds a node.
    This is necessary because ranksame in GraphViz can only be applied to
    individual nodes, not cluster_subgraphs.

    Args:
        step_1_names (List[str]): The list of potential first node names
        yaml_stem (str): The name of the current subworkflow (stem of the yaml filepath)
        namespaces (Namespaces): Specifies the path in the AST of the current subworkflow
        steps_keys (List[str]): The name of each step in the current CWL workflow
        subkeys (List[str]): The keys associated with subworkflows

    Returns:
        str: The name of the first step
    """
    if steps_keys[0] in subkeys:
        step_name_1 = step_1_names[0]
    else:
        step_name_1 = step_name_str(yaml_stem, 0, steps_keys[0])
        step_name_1 = '___'.join(namespaces + [step_name_1])
    # NOTE: Since the names of subgraphs '*.wic' contain a period, we need to
    # escape them by enclosing the whole name in double quotes. Otherwise:
    # "Error: *.wic.gv: syntax error in line n near '.'"
        step_name_1 = f'"{step_name_1}"'

    return step_name_1


def parse_provenance_output_files(output_json: Json) -> List[Tuple[str, str, str]]:
    """Parses the primary workflow provenance JSON object.

    Args:
        output_json (Json): The JSON results object, containing the metadata for all output files.

    Returns:
        List[Tuple[str, str, str]]: A List of (location, parentdirs, basename) for each output file.
    """
    files = []
    for namespaced_output_name, obj in output_json.items():
        files.append(parse_provenance_output_files_(obj, namespaced_output_name))
    return [y for x in files for y in x]


def parse_provenance_output_files_(obj: Any, parentdirs: str) -> List[Tuple[str, str, str]]:
    """Parses the primary workflow provenance JSON object.

    Args:
        obj (Any): The provenance object or one of its recursive sub-objects.
        parentdirs (str): The directory associated with obj.

    Returns:
        List[Tuple[str, str, str]]: A List of (location, parentdirs, basename) for each output file.
    """
    if isinstance(obj, Dict):
        if obj.get('class', '') == 'File':
            return [(str(obj['location']), parentdirs, str(obj['basename']))]  # This basename is a file name
        if obj.get('class', '') == 'Directory':
            subdir = parentdirs + '/' + obj['basename']  # This basename is a directory name
            return parse_provenance_output_files_(obj['listing'], subdir)
    if isinstance(obj, List):
        files = []
        for o in obj:
            files.append(parse_provenance_output_files_(o, parentdirs))
        # Should we flatten?? This will lose the structure of 2D (and higher) array outputs.
        return [y for x in files for y in x]
    return []


def get_input_mappings(input_mapping: Dict[str, List[str]], arg_keys: List[str],
                       arg_key_in_yaml_tree_inputs: bool) -> List[str]:
    """Gets all of the workflow step inputs / call sites that are mapped from the given workflow inputs.

    Args:
        input_mapping (Dict[str, List[str]]): Maps workflow inputs to workflow step inputs, recursively namespaced.
        arg_keys (List[str]): A (singleton) list of root workflow inputs.
        arg_key_in_yaml_tree_inputs (bool): Determines whether at least one level of recursion has been performed.

    Returns:
        List[str]: A list of the workflow step inputs / call sites, recursively namespaced.
    """
    # print('arg_keys', arg_keys)
    # Since each workflow input can be used in many workflow steps, we
    # need to (recursively) find all of the leaves of the mapping tree
    # corresponding to the root arg_key/in_name. Since we already added all
    # sub-input_mapping's (with namespacing) after each recursive call,
    # this flattens the recursion into iteration here. The only trick is
    # that we also need to remove the intermediate variables associated
    # with subworkflow boundaries.
    if not arg_key_in_yaml_tree_inputs:
        done = False
        while not done:
            done = True
            arg_keys_accum = []
            for arg_key_ in arg_keys:
                if arg_key_ in input_mapping:
                    # Remove the intermediate variables associated with subworkflow boundaries.
                    arg_key_init_namespaces = arg_key_.split('___')[:-1]
                    temp = ['___'.join(arg_key_init_namespaces + [s]) for s in input_mapping[arg_key_]]
                    arg_keys_accum.append(temp)
                    done = False
                else:
                    arg_keys_accum.append([arg_key_])
            arg_keys = [y for x in arg_keys_accum for y in x]
            # print('arg_keys', arg_keys)

    return arg_keys


def get_output_mapping(output_mapping: Dict[str, str], out_key: str) -> str:
    """Gets the workflow step output / return location that is mapped to the given workflow output.

    Args:
        output_mapping (Dict[str, str]): Maps workflow outputs to workflow step outputs, recursively namespaced.
        out_key (str): The root workflow output.

    Returns:
        str: The workflow step output / return location, recursively namespaced.
    """
    # print('out_key', out_key)
    # Similarly, we need to find the fixed-point of output_mapping.
    # This is simpler since a workflow output can only come from one workflow step.
    # if not out_key_in_yaml_tree_outputs:
    done = False
    while not done:
        done = True
        # out_key = f'{step_name_j}___{out_key}' # TODO: Check this
        if out_key in output_mapping:
            # Remove the intermediate variables associated with subworkflow boundaries.
            out_key_init_namespaces = out_key.split('___')[:-1]
            out_key = '___'.join(out_key_init_namespaces + [output_mapping[out_key]])
            done = False
        # print('out_key', out_key)

    return out_key



"""CWL Types."""
from enum import Enum
from pathlib import Path
from typing import Any


class CWLTypesEnum(str, Enum):
    NULL = "null"
    BOOLEAN = "boolean"
    INT = "int"
    LONG = "long"
    FLOAT = "float"
    DOUBLE = "double"
    STRING = "string"
    FILE = "File"
    DIRECTORY = "Directory"


class ScatterMethod(Enum):
    dotproduct = "dotproduct"
    flat_crossproduct = "flat_crossproduct"
    nested_crossproduct = "nested_crossproduct"



import argparse
from pathlib import Path
from typing import List

import graphviz
import networkx as nx

from .wic_types import (GraphData, GraphReps, Json, Namespaces, Tool, Tools)


def add_graph_edge(args: argparse.Namespace, graph: GraphReps,
                   nss1: Namespaces, nss2: Namespaces,
                   label: str, color: str = '') -> None:
    """Adds edges to (all of) our graph representations, with the ability to
    collapse all nodes below a given depth to a single node.

    This function utilizes the fact that nodes have been carefully designed to
    have unique, hierarchical names. If we want to hide all of the details
    below a given depth, we can simply truncate each of the namespaces!
    (and do the same when creating the nodes)

    Args:
        args (argparse.Namespace): The command line arguments
        graph (GraphReps): A tuple of a GraphViz DiGraph and a networkx DiGraph
        nss1 (Namespaces): The namespaces associated with the first node
        nss2 (Namespaces): The namespaces associated with the second node
        label (str): The edge label
        color (str, optional): The edge color
    """
    if color == '':
        color = 'black' if args.graph_dark_theme else 'white'
    nss1 = nss1[:(1 + args.graph_inline_depth)]
    edge_node1 = '___'.join(nss1)
    nss2 = nss2[:(1 + args.graph_inline_depth)]
    edge_node2 = '___'.join(nss2)
    graph_gv = graph.graphviz
    graph_nx = graph.networkx
    graphdata = graph.graphdata
    attrs = {}
    # Hide internal self-edges
    if edge_node1 != edge_node2:
        attrs = {'color': color}
        if args.graph_label_edges:
            attrs['label'] = label

        graph_gv.edge(edge_node1, edge_node2, **attrs)
    graph_nx.add_edge(edge_node1, edge_node2)
    graphdata.edges.append((edge_node1, edge_node2, attrs))


def flatten_graphdata(graphdata: GraphData, parent: str = '') -> GraphData:
    """Flattens graphdata by recursively inlineing all subgraphs.

    Args:
        graphdata (GraphData): A data structure which contains recursive subgraphs and other metadata.
        parent (str, optional): The name of the parent graph is encoded into the node attributes so that\n
        the subgraph information can be preserved after flattening. (Used for cytoscape) Defaults to ''.

    Returns:
        GraphData: A GraphDath instance with all of the recursive instances inlined
    """
    subgraphs = [flatten_graphdata(subgraph, str(graphdata.name)) for subgraph in graphdata.subgraphs]

    # NOTE: Even though all of the following default list arguments are [],
    # you MUST explicitly supply the empty lists!!! Otherwise, after
    # instantiation, the lists will contain values from previous instances!!!
    # This shallow copy causes an infinite loop because as we copy nodes,
    # they end up getting appended to the original lists!
    # This makes absolutely no sense. Since the lists are defined at the
    # instance level (NOT the class level), there should be zero sharing!
    g_d = GraphData(str(graphdata.name), [], [], [], [])  # This is fine
    # g_d = GraphData(str(graphdata.name)) # This is NOT fine!
    # i.e. The following statement will NOT print zeros (in the second case)!
    # print(g_d.name, len(g_d.nodes), len(g_d.edges), len(g_d.subgraphs), len(g_d.ranksame))

    for subgraph in subgraphs:
        # We need to add a placeholder node for each subgraph first
        attrs = {} if parent == '' else {'parent': parent}
        # NOTE: This does not yet work with args.graph_inline_depth
        g_d.nodes.append((subgraph.name, attrs))

    for subgraph in subgraphs:
        # Then we can add the nodes and edges from the subgraphs.
        # (Otherwise, cytoscape won't render the subgraphs correctly.)
        for (subnode, subattrs) in subgraph.nodes:
            g_d.nodes.append((subnode, subattrs))
        for (subnode1, subnode2, subattrs) in subgraph.edges:
            g_d.edges.append((subnode1, subnode2, subattrs))

    # Finally, add the nodes and edges from the current graph
    for (node, attrs) in graphdata.nodes:
        attrs['parent'] = graphdata.name
        g_d.nodes.append((node, attrs))
    for (node1, node2, attrs) in graphdata.edges:
        g_d.edges.append((node1, node2, attrs))

    return g_d


def graphdata_to_cytoscape(graphdata: GraphData) -> Json:
    """Converts a flattened graph into cytoscape json format.

    Args:
        graphdata (GraphData): A flattened GraphData instance

    Returns:
        Json: A Json object compatible with cytoscape.
    """
    nodes = []
    for (node, attrs) in list(graphdata.nodes):
        nodes.append({'data': {'id': node, **attrs}})
    edges = []
    for (node1, node2, attrs) in list(graphdata.edges):
        edges.append({'data': {'source': node1, 'target': node2, **attrs}})
    return {'nodes': nodes, 'edges': edges}


def make_tool_dag(tool_stem: str, tool: Tool, graph_dark_theme: bool) -> None:
    """Uses the `dot` executable from the graphviz package to make a Directed
    Acyclic Graph corresponding to the given CWL CommandLineTool

    Args:
        tool_stem (str): The name of the Tool
        tool (Tool): The CWL ComandLineTool
        graph_dark_theme (bool): See args.graph_dark_theme
    """
    (tool_path, tool_cwl) = tool
    yaml_path = f'autogenerated/DAG/{tool_path}'
    Path(yaml_path).parent.mkdir(parents=True, exist_ok=True)
    graph = graphviz.Digraph(name=yaml_path)
    graph.attr(bgcolor="transparent")  # Useful for making slides
    font_edge_color = 'black' if graph_dark_theme else 'white'
    graph.attr(fontcolor=font_edge_color)
    graph.attr(rankdir='LR')
    attrs = {'shape': 'box', 'style': 'rounded, filled'}
    graph.node(tool_stem, fillcolor='lightblue', **attrs)
    for input_cwl in tool_cwl['inputs']:
        input_initial_ns = input_cwl.split('___')[0].split('__')[0]
        input_no_initial_ns = input_cwl.replace(f'{input_initial_ns}__', '')
        # Hide optional inputs that could be confusing.
        if not 'output' in input_no_initial_ns:
            graph.node(f'input_{input_cwl}', label=input_no_initial_ns, fillcolor='lightgreen', **attrs)
            graph.edge(f'input_{input_cwl}', tool_stem, color=font_edge_color)
    for output_cwl in tool_cwl['outputs']:
        output_initial_ns = output_cwl.split('___')[0].split('__')[0]
        output_no_initial_ns = output_cwl.replace(f'{output_initial_ns}__', '')
        graph.node(f'output_{output_cwl}', label=output_no_initial_ns, fillcolor='lightyellow', **attrs)
        graph.edge(tool_stem, f'output_{output_cwl}', color=font_edge_color)
    # NOTE: Since there may be many inputs/outputs and thus edges in complex subworkflows,
    # the layout algorithm for this this .render() call may be very slow! (10+ seconds)
    graph.render(format='png')


def make_plugins_dag(tools: Tools, graph_dark_theme: bool) -> None:
    """Uses the `neato` executable from the graphviz package to make a Directed
    Acyclic Graph consisting of a node for each CWL CommandLineTool and no edges.

    Args:
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        graph_dark_theme (bool): See args.graph_dark_theme
    """
    # NOTE: Do not use the default 'dot' engine. Use neato / fdp / sfdp
    # and set pack=0 to remove the massive blank space around each node.
    # Also note that despite my best efforts, I cannot force graphviz to
    # change the aspect ratio (to double the width for making slides)
    # without simply stretching it and distorting the nodes, so we can just
    # partition the tools into two squares and display them side by side.
    num_tools_half = int(len(list(tools)) / 2)
    for i in [0, 1]:
        yaml_path = f'autogenerated/DAG/plugins{i}'
        Path(yaml_path).mkdir(parents=True, exist_ok=True)
        graph = graphviz.Digraph(name=yaml_path)
        graph.engine = 'neato'
        graph.attr(pack='0')
        graph.attr(bgcolor="transparent")  # Useful for making slides
        font_edge_color = 'black' if graph_dark_theme else 'white'
        graph.attr(fontcolor=font_edge_color)
        for tool in list(tools)[i*num_tools_half:(i+1)*num_tools_half]:
            (tool_path, tool_cwl) = tools[tool]
            attrs = {'shape': 'box', 'style': 'rounded, filled'}
            graph.node(Path(tool_path).stem, fillcolor='lightblue', fontsize="24", width='0.75', **attrs)
        # NOTE: Since there are no edges in this DAG and thus no edge constraints,
        # the layout algorithm for this .render() call is fast. (about 1 second)
        graph.render(format='png')


def add_subgraphs(args: argparse.Namespace,
                  graph: GraphReps,
                  sibling_subgraphs: List[GraphReps],
                  namespaces: Namespaces,
                  step_1_names: List[str],
                  steps_ranksame: List[str]) -> None:
    """Add all subgraphs to the current graph, except for GraphViz subgraphs
    below a given depth, which allows us to hide irrelevant details.

    Args:
        args (argparse.Namespace): The command line arguments
        graph (GraphReps): A tuple of a GraphViz DiGraph and a networkx DiGraph
        sibling_subgraphs (List[Graph]): The subgraphs of the immediate children of the current workflow
        namespaces (Namespaces): Specifies the path in the AST of the current subworkflow
        step_1_names (List[str]): The names of the first step
        steps_ranksame (List[str]): Additional node names to be aligned using ranksame
    """
    graph_gv = graph.graphviz
    graph_nx = graph.networkx
    # Add the cluster subgraphs to the main graph, but we need to add them in
    # reverse order to trick the graphviz layout algorithm.
    for sibling in sibling_subgraphs[::-1]:  # Reverse!
        (sib_graph_gv, sib_graph_nx, sib_graphdata) = sibling
        if len(namespaces) < args.graph_inline_depth:
            graph_gv.subgraph(sib_graph_gv)
        graph_nx.add_nodes_from(sib_graph_nx.nodes)
        graph_nx.add_edges_from(sib_graph_nx.edges)
    for sibling in sibling_subgraphs:
        graph.graphdata.subgraphs.append(sibling.graphdata)
    # Align the cluster subgraphs using the same rank as the first node of each subgraph.
    # See https://stackoverflow.com/questions/6824431/placing-clusters-on-the-same-rank-in-graphviz
    if len(namespaces) < args.graph_inline_depth:
        step_1_names_display = [name for name in step_1_names if len(name.split('___')) < 2 + args.graph_inline_depth]
        if len(step_1_names_display) > 1:
            nodes_same_rank = '\t{rank=same; ' + '; '.join(step_1_names_display) + '}\n'
            graph_gv.body.append(nodes_same_rank)
            graph.graphdata.ranksame = step_1_names_display
        if len(steps_ranksame) > 1:
            nodes_same_rank = '\t{rank=same; ' + '; '.join(steps_ranksame) + '}\n'
            graph_gv.body.append(nodes_same_rank)
            graph.graphdata.ranksame = steps_ranksame


def get_graph_reps(name: str) -> GraphReps:
    """Initialize graph representations

    Args:
        name (str): The name of the graph

    Returns:
        GraphReps: A tuple of graph representations
    """
    graph_gv = graphviz.Digraph(name=f'cluster_{name}')
    graph_gv.attr(newrank='True')
    graph_nx = nx.DiGraph()
    graphdata = GraphData(str(name))
    return GraphReps(graph_gv, graph_nx, graphdata)



from pathlib import Path
import argparse
import copy
import yaml


import uvicorn
from fastapi import FastAPI, Request, status
from fastapi.middleware.cors import CORSMiddleware

from sophios import __version__, compiler
from sophios import run_local, input_output
from sophios.utils_graphs import get_graph_reps
from sophios.utils_yaml import wic_loader
from sophios import utils_cwl
from sophios.cli import get_args
from sophios.wic_types import CompilerInfo, Json, Tool, Tools, StepId, YamlTree, Cwl, NodeData
from sophios.api.utils import converter
# from .auth.auth import authenticate


# helper functions


def remove_dot_dollar(tree: Cwl) -> Cwl:
    """Removes . and $ from dictionary keys, e.g. $namespaces and $schemas. Otherwise, you will get
    {'error': {'statusCode': 500, 'message': 'Internal Server Error'}}
    This is due to MongoDB:
    See https://www.mongodb.com/docs/manual/reference/limits/#Restrictions-on-Field-Names
    Args:
        tree (Cwl): A Cwl document
    Returns:
        Cwl: A Cwl document with . and $ removed from $namespaces and $schemas
    """
    tree_str = str(yaml.dump(tree, sort_keys=False, line_break='\n', indent=2))
    tree_str_no_dd = tree_str.replace('$namespaces', 'namespaces').replace(
        '$schemas', 'schemas').replace('.wic', '_wic')
    tree_no_dd: Cwl = yaml.load(tree_str_no_dd, Loader=wic_loader())  # This effectively copies tree
    return tree_no_dd


def get_yaml_tree(req: Json) -> Json:
    """
    Get the Sophios yaml tree from incoming JSON
    Args:
        req (JSON): A raw JSON content of incoming JSON object
    Returns:
        Cwl: A Cwl document with . and $ removed from $namespaces and $schemas
    """
    wkflw_name = "generic_workflow"
    # args = converter.get_args(wkflw_name)
    # yaml_tree_json: Json = converter.wfb_to_wic(req)
    yaml_tree_json: Json = {}
    return yaml_tree_json


def run_workflow(compiler_info: CompilerInfo, args: argparse.Namespace) -> int:
    """
    Get the Sophios yaml tree from incoming JSON
    Args:
        req (JSON): A raw JSON content of incoming JSON object
    Returns:
        Cwl: A Cwl document with . and $ removed from $namespaces and $schemas
    """
    # ========= WRITE OUT =======================
    input_output.write_to_disk(compiler_info.rose, Path('autogenerated/'), relative_run_path=True)
    # ======== TEST RUN =========================
    retval = run_local.run_local(args, compiler_info.rose, args.cachedir, 'cwltool', False)
    return retval


app = FastAPI()

origins = ["*"]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/", status_code=status.HTTP_200_OK)
# @authenticate
async def root(request: Request) -> Json:
    """The api has 1 route: compile

    Returns:
        Dict[str, str]: {"message": "The api has 1 route: compile"}
    """
    return {"message": "The api has 1 route: compile"}


@app.post("/compile")
# @authenticate
async def compile_wf(request: Request) -> Json:
    """The compile route compiles the json object from http request object built elsewhere

    Args:
        request (Request): request object built elsewhere

    Returns:
        compute_workflow (JSON): workflow json object ready to submit to compute
    """
    print('---------- Compile Workflow! ---------')
    # ========= PROCESS REQUEST OBJECT ==========
    req: Json = await request.json()
    # clean up and convert the incoming object
    req = converter.raw_wfb_to_lean_wfb(req)
    wkflw_name = "generic_workflow"
    args = get_args(wkflw_name)

    workflow_temp = {}
    if req["links"] != []:
        for node in req["nodes"]:
            workflow_temp["id"] = node["id"]
            workflow_temp["step"] = node["cwlScript"]  # Assume dict form
    else:  # A single node workflow
        node = req["nodes"][0]
        workflow_temp = node["cwlScript"]

    workflow_can = utils_cwl.desugar_into_canonical_normal_form(workflow_temp)

    # ========= BUILD WIC COMPILE INPUT =========
    tools_cwl: Tools = {StepId(content["id"], "global"):
                        Tool(".", content["run"]) for content in workflow_can["steps"]}
    # run tag will have the actual CommandLineTool
    wic_obj = {'wic': workflow_can.get('wic', {})}
    plugin_ns = wic_obj['wic'].get('namespace', 'global')

    graph = get_graph_reps(wkflw_name)
    yaml_tree: YamlTree = YamlTree(StepId(wkflw_name, plugin_ns), workflow_can)

    # ========= COMPILE WORKFLOW ================
    compiler_info: CompilerInfo = compiler.compile_workflow(yaml_tree, args, [], [graph], {}, {}, {}, {},
                                                            tools_cwl, True, relative_run_path=True, testing=False)

    # =========== OPTIONAL RUN ==============
    print('---------- Run Workflow locally! ---------')
    retval = run_workflow(compiler_info, args)

    # ======== OUTPUT PROCESSING ================
    # ========= PROCESS COMPILED OBJECT =========
    sub_node_data: NodeData = compiler_info.rose.data
    yaml_stem = sub_node_data.name
    cwl_tree = sub_node_data.compiled_cwl
    yaml_inputs = sub_node_data.workflow_inputs_file
    cwl_tree_no_dd = remove_dot_dollar(cwl_tree)
    yaml_inputs_no_dd = remove_dot_dollar(yaml_inputs)

    # Convert the compiled yaml file to json for labshare Compute.
    cwl_tree_run = copy.deepcopy(cwl_tree_no_dd)
    # for step_key in cwl_tree['steps']:
    #     step_name_i = step_key
    #     step_name_i = step_name_i.replace('.yml', '_yml')  # Due to calling remove_dot_dollar above
    #     # step_name = '__'.join(step_key.split('__')[3:])  # Remove prefix
    #     # Get step CWL from templates
    #     # run_val = next((tval['cwlScript']
    #     #                for _, tval in ict_plugins.items() if step_name == tval['name']), None)
    #     # cwl_tree_run['steps'][step_name_i]['run'] = run_val

    compute_workflow: Json = {}
    compute_workflow = {
        "name": yaml_stem,
        "cwlJobInputs": yaml_inputs_no_dd,
        **cwl_tree_run
    }
    compute_workflow["retval"] = str(retval)
    return compute_workflow


if __name__ == '__main__':
    uvicorn.run(app, host="0.0.0.0", port=3000)


# # ========= PROCESS COMPILED OBJECT =========
    # sub_node_data: NodeData = compiler_info.rose.data
    # yaml_stem = sub_node_data.name
    # cwl_tree = sub_node_data.compiled_cwl
    # yaml_inputs = sub_node_data.workflow_inputs_file

    # # ======== OUTPUT PROCESSING ================
    # cwl_tree_no_dd = remove_dot_dollar(cwl_tree)
    # yaml_inputs_no_dd = remove_dot_dollar(yaml_inputs)

    # # Convert the compiled yaml file to json for labshare Compute.
    # cwl_tree_run = copy.deepcopy(cwl_tree_no_dd)
    # for step_key in cwl_tree['steps']:
    #     step_name_i = step_key
    #     step_name_i = step_name_i.replace('.yml', '_yml')  # Due to calling remove_dot_dollar above
    #     step_name = '__'.join(step_key.split('__')[3:])  # Remove prefix

    #     # Get step CWL from templates
    #     run_val = next((tval['cwlScript']
    #                    for _, tval in ict_plugins.items() if step_name == tval['name']), None)
    #     cwl_tree_run['steps'][step_name_i]['run'] = run_val

    # TODO: set name and driver in workflow builder ui
    # compute_workflow: Json = {}
    # compute_workflow = {
    #     "name": yaml_stem,
    #     "driver": "argo",
    #     # "driver": "cwltool",
    #     "cwlJobInputs": yaml_inputs_no_dd,
    #     **cwl_tree_run
    # }



import sys
from pathlib import Path
import shutil
import subprocess as sub
import traceback
from typing import Dict

import graphviz
import networkx as nx
import yaml

from sophios.utils_yaml import wic_loader
from . import input_output as io
from . import ast, cli, compiler, inference, inlineing, plugins, run_local, utils  # , utils_graphs
from .schemas import wic_schema
from .wic_types import GraphData, GraphReps, Json, StepId, Yaml, YamlTree


def main() -> None:
    """See docs/userguide.md"""
    args = cli.parser.parse_args()
    plugins.logging_filters()

    # User may specify a different homedir
    default_config_file = Path(args.homedir)/'wic'/'global_config.json'
    global_config: Json = io.get_config(Path(args.config_file), default_config_file)

    tools_cwl = plugins.get_tools_cwl(global_config,
                                      args.validate_plugins,
                                      not args.no_skip_dollar_schemas,
                                      args.quiet)
    # This takes ~1 second but it is not really necessary.
    # utils_graphs.make_plugins_dag(tools_cwl, args.graph_dark_theme)
    # pass around config object instead of reading from the disk!
    yml_paths = plugins.get_yml_paths(global_config)

    # Perform initialization via mutating global variables (This is not ideal)
    compiler.inference_rules = global_config.get('inference_rules', {})
    inference.renaming_conventions = global_config.get('renaming_conventions', [])

    # Generate schemas for validation and vscode IntelliSense code completion
    yaml_stems = utils.flatten([list(p) for p in yml_paths.values()])
    schema_store: Dict[str, Json] = {}
    validator = wic_schema.get_validator(tools_cwl, yaml_stems, schema_store, write_to_disk=False)

    # Generating yml schemas every time takes ~20 seconds and guarantees the
    # subworkflow schemas are always up to date. However, since it compiles all
    # yml files, if there are any errors in any of the yml files, the user may
    # be confused by an error message when the --yaml file is correct.
    # For now, require the user to update the schemas manually. In the future,
    # we may use a filewatcher.
    if args.generate_schemas:
        yml_paths_tuples = [(yml_path_str, yml_path)
                            for yml_namespace, yml_paths_dict in yml_paths.items()
                            for yml_path_str, yml_path in yml_paths_dict.items()]

        for yml_path_str, yml_path in yml_paths_tuples:
            schema = wic_schema.compile_workflow_generate_schema(args.homedir, yml_path_str, yml_path,
                                                                 tools_cwl, yml_paths, validator,
                                                                 args.ignore_validation_errors,
                                                                 args.allow_raw_cwl)
            # overwrite placeholders in schema_store. See comment in get_validator()
            schema_store[schema['$id']] = schema

        # Now that we compiled all of the subworkflows once with the permissive/weak schema,
        # compile the root yml workflow again with the restrictive/strict schema.
        validator = wic_schema.get_validator(tools_cwl, yaml_stems, schema_store, write_to_disk=True)

    if args.generate_schemas:
        print('Finished generating schemas. Exiting.')
        sys.exit(0)

    yaml_path = args.yaml
    yaml_stem = Path(args.yaml).stem

    # Load the high-level yaml root workflow file.
    with open(yaml_path, mode='r', encoding='utf-8') as y:
        root_yaml_tree: Yaml = yaml.load(y.read(), Loader=wic_loader())
    Path('autogenerated/').mkdir(parents=True, exist_ok=True)
    wic = {'wic': root_yaml_tree.get('wic', {})}
    plugin_ns = wic['wic'].get('namespace', 'global')
    step_id = StepId(yaml_path, plugin_ns)
    y_t = YamlTree(step_id, root_yaml_tree)
    yaml_tree_raw = ast.read_ast_from_disk(args.homedir, y_t, yml_paths, tools_cwl, validator,
                                           args.ignore_validation_errors)
    # Write the combined workflow (with all subworkflows as children) to disk.
    # NOTE: This is completely optional and for debugging purposes only.
    with open(f'autogenerated/{Path(yaml_path).stem}_tree_raw.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree_raw.yml))
    yaml_tree = ast.merge_yml_trees(yaml_tree_raw, {}, tools_cwl)
    with open(f'autogenerated/{Path(yaml_path).stem}_tree_merged.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))
    root_yml_dir_abs = Path(args.yaml).parent.absolute()
    yaml_tree = ast.python_script_generate_cwl(yaml_tree, root_yml_dir_abs, tools_cwl)
    with open(f'autogenerated/{Path(yaml_path).stem}_tree_python_script.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))

    if args.cwl_inline_subworkflows:
        while True:
            # Inlineing changes the namespaces, so we have to get new namespaces after each inlineing operation.
            namespaces_list = inlineing.get_inlineable_subworkflows(yaml_tree, tools_cwl, False, [])
            if namespaces_list == []:
                break

            # print('inlineing', namespaces_list[0])
            yaml_tree, _len_substeps = inlineing.inline_subworkflow(yaml_tree, namespaces_list[0])

        # Append _inline here instead of in io.write_to_disk()
        step_id = StepId(yaml_tree.step_id.stem + '_inline', yaml_tree.step_id.plugin_ns)
        yaml_tree = YamlTree(step_id, yaml_tree.yml)

        with open(f'autogenerated/{Path(yaml_path).stem}_tree_merged_inlined.wic', mode='w', encoding='utf-8') as f:
            f.write(yaml.dump(yaml_tree.yml))

    rootgraph = graphviz.Digraph(name=yaml_path)
    rootgraph.attr(newrank='True')  # See graphviz layout comment above.
    rootgraph.attr(bgcolor="transparent")  # Useful for making slides
    font_edge_color = 'black' if args.graph_dark_theme else 'white'
    rootgraph.attr(fontcolor=font_edge_color)

    # This can be used to visually 'inline' all subworkflows (but NOT the CWL).
    # rootgraph.attr(style='invis')
    # Note that since invisible objects still affect the graphviz layout (by design),
    # this can be used to control the layout of the individual nodes, even if
    # you don't necessarily want subworkflows.

    # rootgraph.attr(rankdir='LR') # When --graph_inline_depth 1, this usually looks better.
    with rootgraph.subgraph(name=f'cluster_{yaml_path}') as subgraph_gv:
        # get the label (if any) from the workflow
        step_i_wic_graphviz = yaml_tree.yml.get('wic', {}).get('graphviz', {})
        label = step_i_wic_graphviz.get('label', yaml_path)
        subgraph_gv.attr(label=label)
        subgraph_gv.attr(color='lightblue')  # color of cluster subgraph outline
        subgraph_nx = nx.DiGraph()
        graphdata = GraphData(yaml_path)
        subgraph = GraphReps(subgraph_gv, subgraph_nx, graphdata)
        try:
            compiler_info = compiler.compile_workflow(yaml_tree, args, [], [subgraph], {}, {}, {}, {},
                                                      tools_cwl, True, relative_run_path=True, testing=False)
        except Exception as e:
            # Certain constraints are conditionally dependent on values and are
            # not easily encoded in the schema, so catch them here.
            # Moreover, although we check for the existence of input files in
            # stage_input_files, we cannot encode file existence in json schema
            # to check the python_script script: tag before compile time.
            print('Failed to compile', yaml_path)
            print(f'See error_{yaml_stem}.txt for detailed technical information.')
            # Do not display a nasty stack trace to the user; hide it in a file.
            with open(f'error_{yaml_stem}.txt', mode='w', encoding='utf-8') as f:
                # https://mypy.readthedocs.io/en/stable/common_issues.html#python-version-and-system-platform-checks
                if sys.version_info >= (3, 10):
                    traceback.print_exception(type(e), value=e, tb=None, file=f)
                else:
                    traceback.print_exception(etype=type(e), value=e, tb=None, file=f)
            sys.exit(1)
        rose_tree = compiler_info.rose
    rose_tree = plugins.cwl_prepend_dockerFile_include_path_rosetree(rose_tree)

    if args.partial_failure_enable:
        rose_tree = plugins.cwl_update_outputs_optional_rosetree(rose_tree)

    io.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

    # this has to happen after at least one write
    # so we can copy from local cwl_dapters in autogenerated/
    if args.inline_cwl_runtag:
        rose_tree = plugins.cwl_update_inline_runtag_rosetree(rose_tree, Path('autogenerated/'), True)
        io.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

    if args.graphviz:
        if shutil.which('dot'):
            # Render the GraphViz diagram
            rootgraph.save(f'autogenerated/{yaml_stem}.wic.gv')
            cmdline = f'cat autogenerated/{yaml_stem}.wic.gv | dot -Tpng > autogenerated/{yaml_stem}.wic.gv.png'
            sub.run(cmdline, shell=True, capture_output=False, check=False)
            # Use the explicit code above because the .render() method saves
            # the png file in the same directory as the original yaml_path
            # rootgraph.render(format='png')  # Default pdf. See https://graphviz.org/docs/outputs/

            # For comparison, the built-in cwltool graphiz support generates a visual abomination:
            cmdline = f'cwltool_filterlog --print-dot autogenerated/{yaml_stem}.cwl | dot -Tsvg > autogenerated/{yaml_stem}.svg'
            sub.run(cmdline, shell=True, capture_output=True, check=False)
        else:
            print("Warning: Cannot generate graphviz diagrams because the `dot` executable was not found.")
            print("(This may happen if you installed the graphviz python package")
            print("but not the graphviz system package.)")

    if args.run_local or args.generate_run_script:
        # cwl-docker-extract recursively `docker pull`s all images in all subworkflows.
        # This is important because cwltool only uses `docker run` when executing
        # workflows, and if there is a local image available,
        # `docker run` will NOT query the remote repository for the latest image!
        # cwltool has a --force-docker-pull option, but this may cause multiple pulls in parallel.
        if args.container_engine == 'singularity':
            cmd = ['cwl-docker-extract', f'-s --dir {args.homedir}', f'autogenerated/{yaml_stem}.cwl']
        else:
            cmd = ['cwl-docker-extract', '--force-download', f'autogenerated/{yaml_stem}.cwl']
        sub.run(cmd, check=True)

        if args.docker_remove_entrypoints:
            # Requires root, so guard behind CLI option
            if args.container_engine == 'docker':
                plugins.remove_entrypoints_docker()
            if args.container_engine == 'podman':
                plugins.remove_entrypoints_podman()

            rose_tree = plugins.dockerPull_append_noentrypoint_rosetree(rose_tree)
            io.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

        run_local.run_local(args, rose_tree, args.cachedir, args.cwl_runner, False)

        # Finally, since there is an output file copying bug in cwltool,
        # we need to copy the output files manually. See comment above.
        if args.copy_output_files:
            run_local.copy_output_files(yaml_stem)


if __name__ == '__main__':
    main()



import json
from pathlib import Path
import random
from typing import Any, Dict, List, Tuple

import networkx as nx
import graphviz
from jsonschema import RefResolver, Draft202012Validator
import yaml

import sophios
from sophios import ast, compiler, utils_cwl
from sophios.cli import get_args
from sophios.utils_yaml import wic_loader
from sophios.wic_types import GraphData, GraphReps, NodeData, StepId, Yaml, YamlTree
from ..wic_types import Json, Tools


config_schemas = {}
path = Path('../mm-workflows/autogenerated/schemas/config_schemas.json')
if path.exists():
    with open(path, mode='r', encoding='utf-8') as f:
        config_schemas = json.load(f)


def default_schema(url: bool = False) -> Json:
    """A basic default schema (to avoid copy & paste).

    Args:
        url (bool, optional): Determines whether to include the $schema url. Defaults to False.

    Returns:
        Json: A basic default schema
    """
    schema: Json = {}
    schema['type'] = 'object'
    schema['additionalProperties'] = False
    schema['errorMessage'] = 'Property is not allowed. IS YOUR INDENTATION CORRECT?'
    if url:
        schema['$schema'] = 'https://json-schema.org/draft/2020-12/schema'
    return schema


def named_schema(name: str, schema: Json) -> Json:
    """Creates a schema which starts with name and contains the given schema.

    Args:
        name (str): The identifier of the string
        schema (Json): The given schema

    Returns:
        Json: A schema which matches name and the given schema
    """
    schema_ = default_schema()
    schema_['properties'] = {name: schema}
    return schema_


def named_empty_schema(name: str) -> Json:
    """Creates a schema which starts with name, but is otherwise an empty wildcard

    Args:
        name (str): The identifier of the string

    Returns:
        Json: A schema which matches anything starting with name
    """
    return named_schema(name, {})  # NOTE: {} is essentially a wildcard


def named_null_schema(name: str) -> Json:
    """Creates a schema which starts with name and contains nothing else

    Args:
        name (str): The identifier of the string

    Returns:
        Json: A schema which matches name and nothing else
    """
    # NOTE: Use this together with anyOf to allow no explicit arguments
    return named_schema(name, {'type': 'null'})


def named_id_schema(name: str) -> Json:
    """Creates a schema which consists of an 'id' with the given name and nothing else.

    Args:
        name (str): The identifier of the string

    Returns:
        Json: A schema which consists of an 'id' with the given name and nothing else.
    """
    # NOTE: Use this together with anyOf to allow no explicit arguments
    return named_schema('id', {'type': 'string', 'const': name})


def cwl_type_to_jsonschema_type_schema(type_obj: Json) -> Json:
    """Converts a canonicalized CWL type into the equivalent jsonschema type schema, if possible.

    Args:
        type_obj (Json): A canonical CWL type object

    Returns:
        Json: A JSON type schema corresponding to type_obj if valid else None
    """
    jsontype = cwl_type_to_jsonschema_type(type_obj)
    if jsontype is not None:
        if isinstance(jsontype, (List, str)):
            return {'type': jsontype}
    return jsontype


def cwl_type_to_jsonschema_type(type_obj: Json) -> Json:
    """Converts a canonicalized CWL type into the equivalent jsonschema type schema, if possible.

    Args:
        type_obj (Json): A canonical CWL type object

    Returns:
        Json: A JSON type schema corresponding to type_obj if valid else None
    """
    if isinstance(type_obj, str):
        # Obviously ignore empty types
        if type_obj == '':
            return None
        # The 'null' type is valid; it does not indicate an error.
        if type_obj == 'null':
            return 'null'
        # There are only two numeric types in json
        if type_obj == 'int' or type_obj == 'long':
            return 'integer'
        if type_obj == 'float' or type_obj == 'double':
            return 'number'
        # Rename common abbreviations
        if type_obj == 'bool':
            return 'boolean'
        if type_obj == 'str':
            return 'string'
        # CWL supports an Any type; map this to the empty wildcard {} schema
        if type_obj == 'Any':
            return {}
        # json does not have any File or Directory types
        if type_obj == 'File':
            return None
        if type_obj == 'Directory':
            return None
        if type_obj == 'WritableFile':
            return None
        if type_obj == 'WritableDirectory':
            return None

    if isinstance(type_obj, Dict):
        if type_obj.get('type') == 'array' and 'items' in type_obj:
            items = cwl_type_to_jsonschema_type(type_obj['items'])
            if items is None:
                return None  # Propagate any type failures
            if items == {}:  # Explicitly handle the `Any` type
                return {**type_obj, 'items': items}
            if isinstance(type_obj['items'], str) and items != {}:
                # Wrap primitive strings in {'type': ...}
                return {**type_obj, 'items': {'type': items}}
            return {**type_obj, 'items': items}
        # TODO: Other cases?

    if isinstance(type_obj, List):
        items = [cwl_type_to_jsonschema_type(item) for item in type_obj]
        if any([item is None for item in items]):
            return None  # Propagate any type failures
        # See https://cswr.github.io/JsonSchema/spec/multiple_types/
        # In a list, if some of the types are themselves arrays or objects,
        # we need to replace them with "array" and "object". This loses
        # information, but that's the specification, so...
        items = ['array' if isinstance(item, Dict) and item.get('type', '') == 'array' else item for item in items]
        items = ['object' if isinstance(item, Dict) and item.get('type', '') == 'object' else item for item in items]
        return items

    # TODO: Support CWL records
    return type_obj


def cwl_schema(name: str, cwl: Json, id_prefix: str) -> Json:
    """Generates a schema (including documentation) based on the inputs of a CWL CommandLineTool or Workflow.

    Args:
        name (str): The name of the CWL CommandLineTool or Workflow
        cwl (Json): The CWL CommandLineTool or Workflow
        id_prefix (str): Either the string 'tools' or 'workflows'

    Returns:
        Json: An autogenerated, documented schema based on the inputs and outputs of a CWL CommandLineTool or Workflow.
    """
    inputs_props: Json = {}

    str_nonempty = {'type': 'string', 'minLength': 1}
    anytype: Dict[Any, Any] = {}

    # See utils_yaml.py
    alias = default_schema()
    alias['properties'] = {'wic_alias': str_nonempty}  # !*

    ii = default_schema()
    ii['properties'] = {'wic_inline_input': anytype}  # !ii

    # required = []
    for key, val in cwl['inputs'].items():
        metadata = {'title': val.get('label', ''), 'description': val.get('doc', '')}
        str_nonempty = {'type': 'string', 'minLength': 1, **metadata}

        # Determine required keys
        # if key == 'config' or not ('?' in cwltype or 'default' in val):
        #    required.append(key)

        # Handle special cases
        if key == 'config' and name in config_schemas:
            inputs_props[key] = {'anyOf': [str_nonempty, alias, ii,
                                           {**config_schemas[name], **metadata}]}
            continue

        if key == 'config' and name == 'cwl_subinterpreter':
            # This may cause problems with hypothesis, but since the
            # cwl_subinterpreter config tag takes an arbitrary Json-encoded string
            # as input, we cannot restrict this particular sub-schema.
            empty_schema: Dict = {}
            inputs_props[key] = empty_schema
            continue

        if key == 'config' and name == 'config_tag_mdp':
            grompp = config_schemas.get('grompp', {})

            ii_mdp = default_schema()
            ii_mdp['properties'] = {'wic_inline_input': grompp}  # !ii

            inputs_props[key] = ii_mdp
            continue

        # Add type information, with exceptions
        cwltype = utils_cwl.canonicalize_type(val.get('type', ''))
        jsontype = cwl_type_to_jsonschema_type_schema(cwltype)
        if jsontype:
            if jsontype == {'type': 'string'}:
                jsontype = str_nonempty
            if isinstance(jsontype['type'], List) and 'string' in jsontype['type']:
                jsontype['type'].remove('string')
            inputs_props[key] = {'anyOf': [str_nonempty, alias, ii, {**jsontype, **metadata}]}
        else:
            inputs_props[key] = {'anyOf': [str_nonempty, alias, ii]}

    # Do not mark properties which are required for CWL as required for yml,
    # because the whole point of inference is that we shouldn't have to!
    # if not required == []:
    #    inputs_props['required'] = required

    inputs = default_schema()
    inputs['properties'] = inputs_props

    scatter = {}
    if inputs_props:
        # NOTE: The CWL specification defines what happens when a scattered input array is empty at runtime.
        # https://www.commonwl.org/v1.2/Workflow.html#Scatter/gather
        # However, it does not define, in the case that there are no input parameters, if it is
        # permissible to specify the scatter tag with an empty list of parameters.
        # In anticipation of different CWL implementations interpreting this ambiguity differently,
        # let's just banish it here, i.e. only allow the scatter tag when there are inputs.
        consts = [{**val, 'const': key} for key, val in inputs_props.items()]
        scatter_props = {'type': 'array', 'items': {'anyOf': consts}}
        scattermethods = ['dotproduct', 'flat_crossproduct', 'nested_crossproduct']
        scattermethod_props: Json = {'type': 'string', 'enum': scattermethods}
        scatter = {'scatter': scatter_props,
                   'scatterMethod': scattermethod_props}  # NOTE: capital M

    outputs_props: Json = {}
    for key, val in cwl['outputs'].items():
        metadata = {'title': val.get('label', ''), 'description': val.get('doc', '')}
        str_nonempty = {'type': 'string', 'minLength': 1, **metadata}

        # Add type information, with exceptions
        cwltype = utils_cwl.canonicalize_type(val.get('type', ''))
        jsontype = cwl_type_to_jsonschema_type_schema(cwltype)
        if jsontype:
            outputs_props[key] = {'anyOf': [str_nonempty, {**jsontype, **metadata}]}
        else:
            outputs_props[key] = str_nonempty

    outputs = default_schema()
    outputs['properties'] = outputs_props

    # See utils_yaml.py
    anchor = default_schema()
    anchor['properties'] = {'wic_anchor': str_nonempty}  # !&

    keys_anchors: Json = {}
    for key in cwl['outputs'].keys():
        key_schema = default_schema()
        key_schema['properties'] = {key: anchor}

    # NOTE: This function generates schemas compatible with call sites in a
    # workflow. Specifically, the types of `inputs:` and `in:` are the same (Json)
    # but we want to use out, NOT outputs below, which has type List[str].
    outputs_keys = {'type': 'string', 'enum': list(cwl['outputs'].keys())}
    # Moreover, we want to support the wic_anchor !& syntax in the out: tag
    out = {'type': 'array', 'items': {'anyOf': [outputs_keys, keys_anchors]}}

    step_name = name + '.wic' if id_prefix == 'workflows' else name

    step_props = default_schema()
    step_props['title'] = cwl.get('label', '')
    step_props['description'] = cwl.get('doc', '')
    step_props['properties'] = {'in': inputs,
                                'out': out,  # NOT outputs! See comment above!
                                **scatter,
                                # The run: tag can be either a string path to another file,
                                # or the file contents can be inlined. For now, just use {}
                                'run': {},  # TODO
                                'id': {'type': 'string', 'const': step_name},
                                'hints': {},  # TODO
                                'requirements': {},  # TODO
                                }

    # NOTE: See comment in get_validator(). Nonetheless, the vscode YAML extension
    # appears to be resolving ids w.r.t. relative local paths. jsonschema
    # (correctly) treats f'tools/{name}.json' as as uninterpreted string,
    # so instead of using name let's just use fake relative paths in ids.
    step_props['$id'] = f'{id_prefix}/{name}.json'
    return step_props


def wic_tag_schema(hypothesis: bool = False) -> Json:
    """The schema of the (recursive) wic: metadata annotation tag.

    Args:
        hypothesis (bool): Determines whether we should restrict the search space.

    Returns:
        Json: The schema of the (recursive) wic: metadata annotation tag.
    """
    # NOTE: This schema needs to be recursive. Use dynamic anchors / references.
    # See https://json-schema.org/draft/2020-12/json-schema-core.html#dynamic-ref
    # and https://stackoverflow.com/questions/69728686/explanation-of-dynamicref-dynamicanchor-in-json-schema-as-opposed-to-ref-and

    # TODO: restrict the str to the enum of all valid step keys
    pat_int_str = "\\([0-9]+, [A-Za-z0-9_\\.]+\\)"

    graphviz_props: Json = {}
    # Is it useful to have an empty label? Let's bar it for now.
    graphviz_props['label'] = {'type': 'string', 'minLength': 1}
    pat_gv_style = '((,\\s*)*(dashed|dotted|solid|invis|bold|tapered|filled|striped|wedged|diagonals|rounded))+'
    graphviz_props['style'] = {'type': 'string', 'pattern': pat_gv_style}
    graphviz_props['ranksame'] = {'type': 'array'}
    graphviz_props['ranksame']['items'] = {'type': 'string', 'pattern': pat_int_str}

    graphviz_schema = default_schema()
    graphviz_schema['properties'] = graphviz_props

    # Call recursive reference
    recursive_ref = {'$dynamicRef': '#wic'}
    in_props: Json = {}  # TODO: Add yml specific properties
    out_props: Json = {}  # TODO: Add yml specific properties

    scatter_props: Json = {}  # TODO: Add yml specific properties
    scattermethod_props: Json = {'type': 'string', 'enum': ['dotproduct', 'flat_crossproduct', 'nested_crossproduct']}

    choices_props = {'wic': recursive_ref, 'scatterMethod': scattermethod_props}  # NOTE: capital M
    if not hypothesis:
        # Empty wildcard {} schemas can cause problems with hypothesis.
        choices_props['in'] = in_props
        choices_props['out'] = out_props
        choices_props['scatter'] = scatter_props
    choices = default_schema()
    choices['properties'] = choices_props

    # See https://json-schema.org/understanding-json-schema/reference/object.html#patternproperties
    # NOTE: This recursive schema is correct, as determined by jsonschema.validate()
    # However, it seems that the vscode YAML extension does not support recursive
    # schema. (IntelliSense works fine until the first instance of recursion.)
    # TODO: A workaround would be to autogenerate a specific schema for each
    # yml file. We should probably do this anyway for the in: tag.
    steps = default_schema()
    # additionalProperties = False still works with patternProperties FYI
    steps['patternProperties'] = {pat_int_str: choices}

    # implementations = default_schema()
    implementations: Dict[Any, Any] = {}
    implementations['type'] = 'object'
    implementations['additionalProperties'] = True
    # TODO: Restrict the implementation properties and make default_implementation an enum

    str_nonempty = {'type': 'string', 'minLength': 1}

    namespace: Dict[Any, Any] = str_nonempty
    # namespace['enum'] = ...
    # TODO: Restrict the namespace properties to only those in search_paths_wic

    implementation = str_nonempty
    default_implementation = str_nonempty
    inlineable = {'type': 'boolean'}

    schema = default_schema(url=True)
    schema['$id'] = 'wic_tag'
    # Create recursive anchor
    schema['$dynamicAnchor'] = 'wic'
    schema['title'] = 'Metadata annotations'
    schema['description'] = 'Use steps: to recursively overload / pass parameters.\nUse graphviz: to modify the DAGs.'

    pat_semver = "^[0-9]+\\.[0-9]+\\.[0-9]+$"
    version = {'type': 'string', 'pattern': pat_semver}
    driver = {'type': 'string', 'enum': ['slurm', 'argo']}

    schema_props = {'graphviz': graphviz_schema, 'steps': steps, 'implementation': implementation,
                    'default_implementation': default_implementation,
                    'version': str_nonempty, 'driver': driver,
                    'namespace': namespace, 'inlineable': inlineable}
    if not hypothesis:
        # {'additionalProperties': True} can cause problems with hypothesis.
        schema_props['implementations'] = implementations
    schema['properties'] = schema_props
    return schema


def wic_main_schema(tools_cwl: Tools, yml_stems: List[str], schema_store: Dict[str, Json], hypothesis: bool = False) -> Json:
    """The main schema which is used to validate yml files.

    Args:
        tools_cwl (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        yml_stems (List[str]): The names of the yml workflow definitions found using get_yml_paths()
        schema_store (Dict[str, Json]): A global mapping between ids and schemas
        hypothesis (bool): Determines whether we should restrict the search space.

    Returns:
        Json: The main schema which is used to validate yml files.
    """
    wildcard_schema: Json = {}
    wildcard_schema['type'] = 'object'
    wildcard_schema['additionalProperties'] = True

    # NOTE: As mentioned below, using $ref's with external schema files
    # (coincidentally?) works with the VSCode YAML extension, and for the
    # jsonschema library we can supply an explicit schemastore. The API of the
    # hypothesis-jsonschema library, however, only takes a schema. So we either
    # need to bundle the external file contents into wic.json (using $def's),
    # or (since there is only one call site per file) simply inline the contents.
    schemas_tools: List[Tuple[str, Json, Json, Json]] = \
        [(step_id.stem,
          schema_store.get(f'tools/{step_id.stem}.json',
                           {'$ref': f'tools/{step_id.stem}.json'}),
          named_null_schema(step_id.stem),
          named_id_schema(step_id.stem))
         for step_id in tools_cwl if not step_id.stem.startswith('python_script')]
    schemas_tools_list: List[Json] = [{'anyOf': [schema, schema_id]}
                                      for name, schema, schema_null, schema_id in schemas_tools]
    schemas_tools_dict: dict[str, Json] = {name: {'anyOf': [schema, {}]}
                                           for name, schema, schema_null, schema_id in schemas_tools}
#    tools_schemas: List[Json] = [{'anyOf': [{'$ref': f'tools/{step_id.stem}.json'},
#                                            named_null_schema(step_id.stem)]} for step_id in tools_cwl]
    # NOTE: See comment in get_validator(). Nonetheless, the vscode YAML extension
    # appears to be resolving ids w.r.t. relative local paths. jsonschema
    # (correctly) treats f'tools/{name}.json' as an uninterpreted string,
    # so instead of using stem let's just use fake relative paths in ids.

    # NOTE: We could/should re-validate after every AST modification. This will
    # require substantial code changes, so let's not worry about it for now.
    schemas_yml: List[Tuple[str, Json, Json, Json]] = \
        [(yml_stem,
          schema_store.get(f'workflows/{yml_stem}.json',
                           {'$ref': f'workflows/{yml_stem}.json'}),
          named_null_schema(yml_stem),
          named_id_schema(yml_stem))
         for yml_stem in yml_stems]
    schemas_yml_list: List[Json] = [{'anyOf': [schema, schema_id]}
                                    for name, schema, schema_null, schema_id in schemas_yml]
    schemas_yml_dict: dict[str, Json] = {name: {'anyOf': [schema, {}]}
                                         for name, schema, schema_null, schema_id in schemas_yml}
#    yml_schemas: List[Json] = [{'anyOf': [{'$ref': f'workflows/{yml_stem}.json'},
#                                          named_null_schema(f'{yml_stem}.wic')]} for yml_stem in yml_stems]

    steps_list: Json = {}
    steps_list['type'] = 'array'
    steps_list['description'] = 'A list of workflow steps'

    steps_dict: Json = {}
    steps_dict['type'] = 'object'
    steps_dict['description'] = 'An ordered dict of workflow steps'

    steps_dict['properties'] = {**schemas_tools_dict, **schemas_yml_dict}  # **wildcard_schema ??

    if hypothesis:
        # For performance reasons, limit the size of the schema. The first time
        # you call .example(), hypothesis will compile the schema and cache the
        # results. Subsequent .example() calls are nearly instantaneous.
        # The time increases fairly rapidly with k, i.e.
        k = 1  # 1-5 minutes...
        # Choose a random subset so we're not testing the same files
        schemas_tools_list = random.choices(schemas_tools_list, k=k)
        schemas_yml_list = random.choices(schemas_yml_list, k=k)

    steps_schemas_list = schemas_tools_list + schemas_yml_list

    str_nonempty = {'type': 'string', 'minLength': 1}

    if not hypothesis:
        in_schema: Json = {}
        in_schema['type'] = 'object'
        in_schema['additionalProperties'] = True
        in_schema['properties'] = {'script': str_nonempty}

        # See utils_yaml.py
        anchor = default_schema()
        anchor['properties'] = {'wic_anchor': str_nonempty}  # !&

        # NOTE: We do not know the specific keys statically, so we have to use str_nonempty
        out_schema: Json = {'type': 'array', 'items': {'anyOf': [str_nonempty, anchor]}}

        in_out_schema = default_schema()
        in_out_schema['properties'] = {'in': in_schema, 'out': out_schema}

        python_script_schema = default_schema()
        python_script_schema['properties'] = {'python_script': in_schema}

        steps_schemas_list += [python_script_schema]

    if not hypothesis:
        # In tools_schemas and yml_schemas above, we are able to use the step name
        # to uniquely identify the associated schema of the child node.
        # We want to support raw CWL, where the step name is not necessarily
        # the name of the CommandLineTool / Subworkflow.
        # We still know which schema we should use by looking at the run tag.
        # This shouldn't be terribly difficult to implement, but for now
        # let's just use a wildcard schema with arbitrary keys and values.
        # (So for now intellisense code completion is unavailable for these steps.)
        # Crucially, code completion *is preserved* for tools_schemas and yml_schemas!
        steps_schemas_list += [wildcard_schema]

    steps_list['items'] = {'anyOf': steps_schemas_list, 'minItems': 1, 'title': 'Valid workflow steps'}
    if hypothesis:
        # For performance reasons, limit the number of steps.
        # This should (hopefully) avoid hypothesis.errors.DeadlineExceeded
        steps_list['items']['maxItems'] = 5

    # TODO: Use the real CWL inputs schema
    inputs: Dict[Any, Any] = {}
    inputs['type'] = 'object'
    inputs['additionalProperties'] = True

    # TODO: Use the real CWL outputs schema
    outputs: Dict[Any, Any] = {}
    outputs['type'] = 'object'
    outputs['additionalProperties'] = True

    schema = default_schema(url=True)
    schema['$id'] = 'wic_main'
    schema['title'] = 'Validating against the Workflow Inference Compiler schema'
    # schema['description'] = ''
    # schema['required'] = ['steps'] # steps are not required, e.g. npt.wic

    steps = {'anyOf': [steps_list, steps_dict]}

    schema_props = {'steps': steps,
                    'class': str_nonempty,
                    'cwlVersion': str_nonempty,  # TODO enum https://www.commonwl.org/v1.2/Workflow.html#CWLVersion
                    # TODO https://www.commonwl.org/v1.2/SchemaSalad.html#Explicit_context
                    '$base': str_nonempty,
                    '$namespaces': {},  # TODO
                    '$schemas': {},  # TODO
                    '$graph': {},  # TODO
                    '$import': {},  # TODO https://www.commonwl.org/v1.2/SchemaSalad.html#Import
                    '$include': {},  # TODO https://www.commonwl.org/v1.2/SchemaSalad.html#Include
                    'id': str_nonempty,
                    'hints': {},  # TODO
                    'requirements': {},  # TODO
                    'label': str_nonempty,
                    'doc': str_nonempty}
    # schema_props['wic'] = wic_tag_schema(hypothesis) # NOTE: This technically 'works'
    # with hypothesis, but the wic_tag_schema still needs some work.

    if not hypothesis:
        schema_props['wic'] = wic_tag_schema(hypothesis)
        # {'additionalProperties': True} can cause problems with hypothesis.
        schema_props['inputs'] = inputs
        schema_props['outputs'] = outputs

    schema['properties'] = schema_props

    # https://json-schema.org/understanding-json-schema/structuring.html#bundling
    # import copy
    # schema['$defs'] = copy.deepcopy(schema_store)
    # Without deepcopy, "ValueError: Circular reference detected"
    # "f.write(json.dumps(schema, indent=2))"

    return schema


def compile_workflow_generate_schema(homedir: str,
                                     yml_path_str: str, yml_path: Path,
                                     tools_cwl: Tools,
                                     yml_paths: Dict[str, Dict[str, Path]],
                                     validator: Draft202012Validator,
                                     ignore_validation_errors: bool,
                                     allow_raw_cwl: bool) -> Json:
    """Compiles a workflow and generates a schema which (recursively) includes the inputs/outputs from subworkflows.

    Args:
        homedir (str): The users home directory
        yml_path_str (str): The stem of the path to the yml file
        yml_path (Path): The path to the yml file
        tools_cwl (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        yml_paths (Dict[str, Dict[str, Path]]): The yml workflow definitions found using get_yml_paths()
        validator (Draft202012Validator): Used to validate the yml files against the autogenerated schema.
        ignore_validation_errors (bool): Temporarily ignore validation errors. Do not use this permanently!
        allow_raw_cwl (bool): Do not check whether the input to a workflow step refers to the workflow inputs: tag

    Returns:
        Json: An autogenerated, documented schema based on the inputs and outputs of the Workflow.
    """
    # First compile the workflow.
    # Load the high-level yaml workflow file.
    with open(yml_path, mode='r', encoding='utf-8') as y:
        root_yaml_tree: Yaml = yaml.load(y.read(), Loader=wic_loader())
    Path('autogenerated/').mkdir(parents=True, exist_ok=True)
    wic_tag = {'wic': root_yaml_tree.get('wic', {})}
    plugin_ns = wic_tag['wic'].get('namespace', 'global')
    step_id = StepId(yml_path_str, plugin_ns)
    y_t = YamlTree(step_id, root_yaml_tree)
    yaml_tree_raw = sophios.ast.read_ast_from_disk(homedir, y_t, yml_paths, tools_cwl, validator,
                                                   ignore_validation_errors)
    # with open(f'autogenerated/{Path(yml_path).stem}_tree_raw.wic', mode='w', encoding='utf-8') as f:
    #    f.write(yaml.dump(yaml_tree_raw.yml))
    yaml_tree = sophios.ast.merge_yml_trees(yaml_tree_raw, {}, tools_cwl)
    # with open(f'autogenerated/{Path(yml_path).stem}_tree_merged.wic', mode='w', encoding='utf-8') as f:
    #    f.write(yaml.dump(yaml_tree.yml))
    root_yml_dir_abs = Path(yml_path).parent.absolute()
    yaml_tree = ast.python_script_generate_cwl(yaml_tree, root_yml_dir_abs, tools_cwl)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_python_script.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))

    graph_gv = graphviz.Digraph(name=f'cluster_{yml_path}')
    graph_gv.attr(newrank='True')
    graph_nx = nx.DiGraph()
    graphdata = GraphData(str(yml_path))
    graph = GraphReps(graph_gv, graph_nx, graphdata)
    args = get_args(str(yml_path), ['--allow_raw_cwl'] if allow_raw_cwl else [])
    compiler_info = sophios.compiler.compile_workflow(yaml_tree, args, [], [graph], {}, {}, {}, {},
                                                      tools_cwl, True, relative_run_path=True, testing=True)
    rose_tree = compiler_info.rose
    sub_node_data: NodeData = rose_tree.data

    # wic.io.write_to_disk(rose_tree, Path('autogenerated/'), relative_run_path=True)
    schema = cwl_schema(step_id.stem, sub_node_data.compiled_cwl, 'workflows')

    # with open(f'autogenerated/schemas/workflows/{step_id.stem}.json', mode='w', encoding='utf-8') as f:
    #    f.write(json.dumps(schema, indent=2))

    return schema


def get_validator(tools_cwl: Tools, yml_stems: List[str], schema_store: Dict[str, Json] = {},
                  write_to_disk: bool = False, hypothesis: bool = False) -> Draft202012Validator:
    """Generates the main schema used to check the yml files for correctness and returns a validator.

    Args:
        tools_cwl (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        yml_stems (List[str]): The names of the yml workflow definitions found using get_yml_paths()
        schema_store (Dict[str, Json]): A global mapping between ids and schemas
        write_to_disk (bool): Controls whether to write the schemas to disk.
        hypothesis (bool): Determines whether we should restrict the search space.

    Returns:
        Draft202012Validator: A validator which is used to check the yml files for correctness.
    """
    for step_id, tool in tools_cwl.items():
        schema_tool = cwl_schema(step_id.stem, tool.cwl, 'tools')
        schema_store[schema_tool['$id']] = schema_tool
        # if write_to_disk:
        #    with open(f'autogenerated/schemas/tools/{step_id.stem}.json', mode='w', encoding='utf-8') as f:
        #        f.write(json.dumps(schema_tool, indent=2))

    # Add temporary placeholders to the schema_store so we don't get
    # "jsonschema.exceptions.RefResolutionError: unknown url type: 'workflows/*.json'"
    for yml_stem in yml_stems:
        if f'workflows/{yml_stem}.json' not in schema_store:
            schema_store[f'workflows/{yml_stem}.json'] = {}

    schema = wic_main_schema(tools_cwl, yml_stems, schema_store, hypothesis)
    schema_store[schema['$id']] = schema
    schema_store['wic_tag'] = wic_tag_schema(hypothesis)
    if write_to_disk:
        with open('autogenerated/schemas/wic.json', mode='w', encoding='utf-8') as f:
            f.write(json.dumps(schema, indent=2))

    # Load cached schema from disk
    # NOTE: This may or may not be the correct solution. We should double check
    # all of the call sites of get_validator and the write_to_disk parameters.
    if Path('autogenerated/schemas/wic.json').exists():
        with open('autogenerated/schemas/wic.json', mode='r', encoding='utf-8') as r:
            schema = json.loads(r.read())

    # See https://stackoverflow.com/questions/53968770/how-to-set-up-local-file-references-in-python-jsonschema-document
    # The $ref tag refers to URIs defined in $id tags, NOT relative paths on
    # the local filesystem! We need to create a global mapping between ids and schemas
    # i.e. schema_store.
    resolver = RefResolver.from_schema(schema, store=schema_store)
    """ Use check_schema to 'first verify that the provided schema is
    itself valid, since not doing so can lead to less obvious error
    messages and fail in less obvious or consistent ways.'
    """
    # i.e. This should match 'https://json-schema.org/draft/2020-12/schema'
    # NOTE: If you get nasty errors while developing the schema such as:
    # "jsonschema.exceptions.SchemaError: ... is not valid under any of the given schemas"
    # try temporarily commenting this line out to generate the schema anyway.
    # Then, in any yml file, the very first line should show a "schema stack trace"
    Draft202012Validator.check_schema(schema)
    validator = Draft202012Validator(schema, resolver=resolver)
    return validator



import copy
import logging
import json
import glob
import os
from pathlib import Path
import re
import sys
import tempfile
import traceback
from typing import Any, Dict, Union

import cwltool.load_tool
import yaml
import podman
import docker


from . import input_output as io, utils_cwl, cli
from .python_cwl_adapter import import_python_file
from .wic_types import Cwl, NodeData, RoseTree, StepId, Tool, Tools, Json


# Filter out the "... previously defined" id uniqueness validation warnings
# from line 1162 of ref_resolver.py in the schema_salad library.
# TODO: Figure out if there is a problem with our autogenerated CWL.
class NoPreviouslyDefinedFilter(logging.Filter):
    # pylint:disable=too-few-public-methods
    def filter(self, record: logging.LogRecord) -> bool:
        return not record.getMessage().endswith('previously defined')


class NoResolvedFilter(logging.Filter):
    # pylint:disable=too-few-public-methods
    def filter(self, record: logging.LogRecord) -> bool:
        m = re.match(r"Resolved '.*' to '.*'", record.getMessage())
        return not bool(m)  # (True if m else False)


class NoPartialFailureNullWarning(logging.Filter):
    # pylint:disable=too-few-public-methods
    def filter(self, record: logging.LogRecord) -> bool:
        err_str = "Source is from conditional step and may produce `null`"
        # this will indiscriminately filter this error string
        return not err_str in record.getMessage()


def logging_filters(allow_pf: bool = False) -> None:
    logger_salad = logging.getLogger("salad")
    logger_salad.addFilter(NoPreviouslyDefinedFilter())
    logger_cwltool = logging.getLogger("cwltool")
    logger_cwltool.addFilter(NoResolvedFilter())
    if allow_pf:
        logger_cwltool.addFilter(NoPartialFailureNullWarning())


logger_wicad = logging.getLogger("wicautodiscovery")


def validate_cwl(cwl_path_str: str, skip_schemas: bool) -> None:
    """This is the body of cwltool.load_tool.load_tool but exposes skip_schemas for performance.
    Skipping significantly improves initial validation performance, but this is not always desired.
    See https://github.com/common-workflow-language/cwltool/issues/623

    Args:
        cwl_path_str (str): The path to the CWL file.
        skip_schemas (bool): Skips processing $schemas tags.
    """
    # NOTE: This uses NoResolvedFilter to suppress the info messages to stdout.
    loading_context, workflowobj, uri = cwltool.load_tool.fetch_document(cwl_path_str)
    # NOTE: There has been a breaking change in the API for skip_schemas.
    # TODO: re-enable skip_schemas while satisfying mypy
    # loading_context.skip_schemas = skip_schemas
    loading_context, uri = cwltool.load_tool.resolve_and_validate_document(
        loading_context, workflowobj, uri, preprocess_only=False  # , skip_schemas=skip_schemas
    )
    # NOTE: Although resolve_and_validate_document does some validation,
    # some additional validation is done in make_tool, i.e.
    # resolve_and_validate_document does not in fact throw an exception for
    # some invalid CWL files, but make_tool does!
    process_ = cwltool.load_tool.make_tool(uri, loading_context)
    # return process_ # ignore process_ for now


def get_tools_cwl(config: Json, validate_plugins: bool = False,
                  skip_schemas: bool = False, quiet: bool = False) -> Tools:
    """Uses glob() to find all of the CWL CommandLineTool definition files within any subdirectory of cwl_dir

    Args:
        config_file (Json): The user specified (or default generated) config json object
        cwl_dirs_file (Path): The subdirectories in which to search for CWL CommandLineTools
        validate_plugins (bool, optional): Performs validation on all CWL CommandLiineTools. Defaults to False.
        skip_schemas (bool, optional): Skips processing $schemas tags. Defaults to False.
        quiet (bool, optional): Determines whether it captures stdout or stderr. Defaults to False.

    Returns:
        Tools: The CWL CommandLineTool definitions found using glob()
    """
    cwl_dirs_tag = config['search_paths_cwl']
    # Load ALL of the tools.
    tools_cwl: Tools = {}
    for plugin_ns in cwl_dirs_tag:
        cwl_dirs = cwl_dirs_tag[plugin_ns]
        for cwl_dir in cwl_dirs:
            # "PurePath.relative_to() requires self to be the subpath of the argument, but os.path.relpath() does not."
            # See https://docs.python.org/3/library/pathlib.html#id4 and
            # See https://stackoverflow.com/questions/67452690/pathlib-path-relative-to-vs-os-path-relpath
            pattern_cwl = str(Path(cwl_dir) / '**/*.cwl')
            # print(pattern_cwl)
            cwl_paths = glob.glob(pattern_cwl, recursive=True)
            Path('autogenerated/schemas/tools/').mkdir(parents=True, exist_ok=True)
            if len(cwl_paths) == 0:
                print(f'Warning! No cwl files found in {cwl_dir}.')
                print("Check 'search_paths_cwl' in your config file")
                print('This almost certainly means you are not in the correct directory.')

            for cwl_path_str in cwl_paths:
                if 'biobb_md' in cwl_path_str:
                    continue  # biobb_md is deprecated (in favor of biobb_gromacs)
                # print(cwl_path)
                with open(cwl_path_str, mode='r', encoding='utf-8') as f:
                    tool: Cwl = yaml.safe_load(f.read())
                stem = Path(cwl_path_str).stem
                # print(stem)

                if validate_plugins:
                    validate_cwl(cwl_path_str, skip_schemas)

                tool = utils_cwl.desugar_into_canonical_normal_form(tool)

                if quiet:
                    # Capture stdout and stderr
                    if not 'stdout' in tool:
                        tool.update({'stdout': f'{stem}.out'})
                    if not 'stderr' in tool:
                        tool.update({'stderr': f'{stem}.err'})
                cwl_path_abs = os.path.abspath(cwl_path_str)
                tools_cwl[StepId(stem, plugin_ns)] = Tool(cwl_path_abs, tool)
                # print(tool)
                # utils_graphs.make_tool_dag(stem, (cwl_path_str, tool))
    return tools_cwl


def cwl_update_outputs_optional(cwl: Cwl) -> Cwl:
    """Updates outputs as optional (if any)

    Args:
        cwl (Cwl): A CWL CommandLineTool

    Returns:
        Cwl: A CWL CommandLineTool with outputs optional (if any)
    """
    cwl_mod = copy.deepcopy(cwl)
    # Update success codes to allow simple failures
    args = cli.parser.parse_args()
    # default value of cwl_mod['successCodes'] = 0 and 'partial_failure_success_codes' = [0,1]
    # take the Union of 0 and partial_failure_success_codes, the 'successCodes' might not be set
    # never discard 0 (actual success)
    failure_code_range = args.partial_failure_success_codes_range
    codes_from_range = list(range(failure_code_range[0], failure_code_range[1]))
    direct_failure_codes = args.partial_failure_success_codes
    assert failure_code_range[0] <= failure_code_range[1], \
        f"lower {failure_code_range[0]}  value can't be greater than higher {failure_code_range[1]} value"
    cwl_mod['successCodes'] = list(set([0] + direct_failure_codes + codes_from_range))
    # Update outputs optional
    for out_key, out_val_dict in cwl_mod['outputs'].items():
        if isinstance(out_val_dict['type'], str) and out_val_dict['type'][-1] != '?':
            out_val_dict['type'] += '?'
    return cwl_mod


def cwl_update_inline_runtag(cwl: Cwl, path: Path, relative_run_path: bool) -> Cwl:
    """Updates 'run' tag with inline content

    Args:
        cwl (Cwl): A CWL Workflow
        path (Path): The directory in which to read files from
        relative_run_path (bool): Controls whether to use subdirectories or just one directory
    Returns:
        Cwl: A CWL Workflow with inline run (if any)
    """
    cwl_mod = copy.deepcopy(cwl)
    for step in cwl_mod['steps']:
        runtag_orig = step.get('run', '')
        if isinstance(runtag_orig, str) and runtag_orig.endswith('.cwl'):
            if relative_run_path:
                yml_path = Path.cwd() / path / runtag_orig
            else:
                yml_path = Path(runtag_orig)  # Assume absolute path in the runtag
            with open(yml_path, mode='r', encoding='utf-8') as f:
                runtag_raw = yaml.safe_load(f.read())
                step['run'] = runtag_raw
        else:
            pass  # We only care if the runtag is a cwl filepath
    return cwl_mod


Client = Union[docker.DockerClient, podman.PodmanClient]  # type: ignore


def remove_entrypoints(client: Client, build: Any) -> None:
    """remove/overwrite the entrypoints from all images

    https://cwl.discourse.group/t/override-docker-entrypoint-in-command-line-tool/695/2
    "Here is what the CWL standards have to say about software container entrypoints"
    "I recommend changing your docker container to not use ENTRYPOINT."

    This script will remove / overwrite the entrypoints in ALL of the docker images on your local machine.
    (It will append -noentrypoint to the version tag.)

    The reason is that we want to simultaneously allow
    1. release environments, where users will simply run code in a docker image and
    2. dev/test environments, where developers can run the latest code on the host machine and/or in the CI.

    With entrypoints, there is no easy way to switch between 1 and 2; developers will
    have to manually prepend the entrypoint string to the baseCommand, and/or possibly
    modify paths to be w.r.t. their host machine instead of w.r.t. the image. (/opt/.../main.py)

    Without entrypoints, to switch between 1 and 2, simply comment out DockerRequirement ... that's it!
    The point is that we want a uniform API so that we can programmatically switch between 1 and 2 in the CI.
    We want to run the integration tests first, and only push releases to dockerhub when the tests pass (NOT vice versa!).
    """
    # Get a list of all Docker images
    images = client.images.list()

    # Iterate over each container
    for image in images:
        if len(image.attrs['RepoTags']) != 0 and len(image.tags) != 0:
            for tag in image.tags:
                if tag.endswith('-noentrypoint'):
                    continue

                # Define the content of the Dockerfile
                dockerfile_content = f'''
                FROM {tag}
                ENTRYPOINT []
                '''
                with tempfile.TemporaryDirectory() as tempdir:
                    with open(tempdir + '/Dockerfile_tmp', mode='w', encoding='utf-8') as f:
                        f.write(dockerfile_content)

                    # Build the new Docker image from the Dockerfile
                    new_image, build_logs = build.build(
                        path=tempdir,
                        dockerfile="Dockerfile_tmp",
                        tag=f"{tag}-noentrypoint"
                    )


def remove_entrypoints_docker() -> None:
    """remove/overwrite the entrypoints from all images (using docker)
    """
    client = docker.from_env()  # type: ignore
    remove_entrypoints(client, client.images)


def remove_entrypoints_podman() -> None:
    """remove/overwrite the entrypoints from all images (using podman)
    """
    # See https://github.com/containers/podman-py?tab=readme-ov-file#example-usage
    uri = "unix:///run/user/1000/podman/podman.sock"

    with podman.PodmanClient(base_url=uri) as client:
        remove_entrypoints(client, podman.domain.images_build.BuildMixin())


def cwl_update_inline_runtag_rosetree(rose_tree: RoseTree, path: Path, relative_run_path: bool) -> RoseTree:
    """Inlines the compiled CWL files runtag

    Args:
        rose_tree (RoseTree): The data associated with compiled subworkflows
        path (Path): The directory in which to read files from
        relative_run_path (bool): Controls whether to use subdirectories or just one directory.
    Returns:
        RoseTree: rose_tree with inline cwl runtag
    """
    n_d: NodeData = rose_tree.data
    if n_d.compiled_cwl['class'] == 'Workflow':
        outputs_inline_cwl_runtag = cwl_update_inline_runtag(n_d.compiled_cwl, path, relative_run_path)
    else:
        outputs_inline_cwl_runtag = n_d.compiled_cwl

    sub_trees_path = [cwl_update_inline_runtag_rosetree(sub_rose_tree, path, relative_run_path) for
                      sub_rose_tree in rose_tree.sub_trees]
    node_data_path = NodeData(n_d.namespaces, n_d.name, n_d.yml, outputs_inline_cwl_runtag, n_d.tool,
                              n_d.workflow_inputs_file, n_d.explicit_edge_defs, n_d.explicit_edge_calls,
                              n_d.graph, n_d.inputs_workflow, n_d.step_name_1)
    return RoseTree(node_data_path, sub_trees_path)


def cwl_update_outputs_optional_rosetree(rose_tree: RoseTree) -> RoseTree:
    """Updates outputs optional for every CWL CommandLineTool

    Args:
        rose_tree (RoseTree): The RoseTree returned from compile_workflow(...).rose_tree

    Returns:
        RoseTree: rose_tree with output optional updates to every CWL CommandLineTool
    """
    n_d: NodeData = rose_tree.data
    if n_d.compiled_cwl['class'] == 'CommandLineTool':
        outputs_optional_cwl = cwl_update_outputs_optional(n_d.compiled_cwl)
    else:
        outputs_optional_cwl = n_d.compiled_cwl

    sub_trees_path = [cwl_update_outputs_optional_rosetree(sub_rose_tree) for
                      sub_rose_tree in rose_tree.sub_trees]
    node_data_path = NodeData(n_d.namespaces, n_d.name, n_d.yml, outputs_optional_cwl, n_d.tool,
                              n_d.workflow_inputs_file, n_d.explicit_edge_defs, n_d.explicit_edge_calls,
                              n_d.graph, n_d.inputs_workflow, n_d.step_name_1)
    return RoseTree(node_data_path, sub_trees_path)


def dockerPull_append_noentrypoint(cwl: Cwl) -> Cwl:
    """Appends -noentrypoint to the dockerPull version tag (if any)

    Args:
        cwl (Cwl): A CWL CommandLineTool

    Returns:
        Cwl: A CWL CommandLineTool, with -noentrypoint appended to the dockerPull version tag (if any)
    """
    docker_image: str = cwl.get('requirements', {}).get('DockerRequirement', {}).get('dockerPull', '')
    if docker_image:
        print('docker_image', docker_image)
    if ':' in docker_image:
        repo, tag = docker_image.split(':')
    else:
        repo, tag = docker_image, 'latest'
    if repo and tag and not tag.endswith('-noentrypoint'):
        print('repo, tag', repo, tag)
        image_noentrypoint = repo + ':' + tag + '-noentrypoint'
        cwl_noentrypoint = copy.deepcopy(cwl)
        cwl_noentrypoint['requirements']['DockerRequirement']['dockerPull'] = image_noentrypoint
        return cwl_noentrypoint
    else:
        return cwl


def dockerPull_append_noentrypoint_tools(tools: Tools) -> Tools:
    """Appends -noentrypoint to the dockerPull version tag for every tool in tools.

    Args:
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()

    Returns:
        Tools: tools with -noentrypoint appended to all of the dockerPull version tags.
    """
    return {stepid: Tool(tool.run_path, dockerPull_append_noentrypoint(tool.cwl))
            for stepid, tool in tools.items()}


def dockerPull_append_noentrypoint_rosetree(rose_tree: RoseTree) -> RoseTree:
    """Appends -noentrypoint to the dockerPull version tag for every CWL CommandLineTool

    Args:
        rose_tree (RoseTree): The RoseTree returned from compile_workflow(...).rose_tree

    Returns:
        RoseTree: rose_tree with -noentrypoint appended to the dockerPull version tag for every CWL CommandLineTool
    """
    n_d: NodeData = rose_tree.data
    # NOTE: Since only class: CommandLineTool should have dockerPull tags,
    # this should be the identity function on class: Workflow.
    compiled_cwl_noent = dockerPull_append_noentrypoint(n_d.compiled_cwl)

    sub_trees_noent = [dockerPull_append_noentrypoint_rosetree(sub_rose_tree) for sub_rose_tree in rose_tree.sub_trees]
    node_data_noent = NodeData(n_d.namespaces, n_d.name, n_d.yml, compiled_cwl_noent, n_d.tool, n_d.workflow_inputs_file,
                               n_d.explicit_edge_defs, n_d.explicit_edge_calls, n_d.graph, n_d.inputs_workflow,
                               n_d.step_name_1)
    return RoseTree(node_data_noent, sub_trees_noent)


def cwl_prepend_dockerFile_include_path(cwl: Cwl, cwl_path: str) -> Cwl:
    """Prepends (original) cwl_path to the dockerFile $include path

    Args:
        cwl (Cwl): A CWL CommandLineTool
        cwl_path : Specifies the absolute path of .cwl file
    Returns:
        Cwl: A CWL CommandLineTool, with correct filepath prepended
    """
    cwl_mod = copy.deepcopy(cwl)
    inc_path: str = cwl.get('hints', {}).get('DockerRequirement', {}).get('dockerFile', {}).get('$include', '')
    if inc_path:
        # cwl_path is an absolute path including (original) filename .cwl
        # we just need to prepend the dir of cwl_path to the inc_path
        cwl_dir: str = str(Path(cwl_path).parent)
        cwl_mod['hints']['DockerRequirement']['dockerFile']['$include'] = cwl_dir + '/' + inc_path
    return cwl_mod


def cwl_prepend_dockerFile_include_path_rosetree(rose_tree: RoseTree) -> RoseTree:
    """Prepends (original) cwl_path to the dockerFile $include path to every CWL CommandLineTool

    Args:
        rose_tree (RoseTree): The RoseTree returned from compile_workflow(...).rose_tree

    Returns:
        RoseTree: rose_tree with (original) cwl_path to the dockerFile to the dockerFile $include path
        to every CWL CommandLineTool
    """
    n_d: NodeData = rose_tree.data
    # NOTE: Since only class: CommandLineTool should have dockerPull tags,
    # this should be the identity function on class: Workflow.
    prepended_cwl = cwl_prepend_dockerFile_include_path(n_d.compiled_cwl, n_d.tool.run_path)

    sub_trees_path = [cwl_prepend_dockerFile_include_path_rosetree(sub_rose_tree) for
                      sub_rose_tree in rose_tree.sub_trees]
    node_data_path = NodeData(n_d.namespaces, n_d.name, n_d.yml, prepended_cwl, n_d.tool, n_d.workflow_inputs_file,
                              n_d.explicit_edge_defs, n_d.explicit_edge_calls, n_d.graph, n_d.inputs_workflow,
                              n_d.step_name_1)
    return RoseTree(node_data_path, sub_trees_path)


def get_workflow_paths(config: Json, extension: str) -> Dict[str, Dict[str, Path]]:
    """Uses glob() to recursively find all of the yml workflow definition files
    within any subdirectory of each yml_dir in search_paths_wic.
    NOTE: This function assumes all yml files found are workflow definition files,
    so do not mix regular yml files and workflow files in the same root directory.
    Moreover, each yml_dir should be disjoint; do not use both '.' and './subdir'!

    Args:
        config (Json): The user specified (or default generated) config json object
        extension (str): The filename extension (either 'wic' or 'py')

    Returns:
        Dict[str, Dict[str, Path]]: A dict containing the filepath stem and filepath of each yml file
    """
    search_paths_wic_tag = config['search_paths_wic']
    # Glob all of the yml files too, so we don't have to deal with relative paths.
    yml_paths_all: Dict[str, Dict[str, Path]] = {}
    for yml_namespace in search_paths_wic_tag:
        yml_dirs = search_paths_wic_tag[yml_namespace]
        for yml_dir in yml_dirs:
            # "PurePath.relative_to() requires self to be the subpath of the argument, but os.path.relpath() does not."
            # See https://docs.python.org/3/library/pathlib.html#id4 and
            # See https://stackoverflow.com/questions/67452690/pathlib-path-relative-to-vs-os-path-relpath
            pattern_yml = str(Path(yml_dir) / f'**/*.{extension}')
            yml_paths_sorted = sorted(glob.glob(pattern_yml, recursive=True), key=len, reverse=True)
            Path('autogenerated/schemas/workflows/').mkdir(parents=True, exist_ok=True)
            if len(yml_paths_sorted) == 0:
                print(f'Warning! No {extension} files found in {yml_dir}.')
                print("Check 'search_paths_wic' in your config file")
                print('This almost certainly means you are not in the correct directory.')
            yml_paths = {}
            for yml_path_str in yml_paths_sorted:
                # Exclude our autogenerated inputs files
                if '_inputs' not in yml_path_str:
                    yml_path = Path(yml_path_str)
                    yml_path_abs = os.path.abspath(yml_path_str)
                    yml_paths[yml_path.stem] = Path(yml_path_abs)
            # Check for existing entry
            ns_dict = yml_paths_all.get(yml_namespace, {})
            yml_paths_all[yml_namespace] = {**ns_dict, **yml_paths}

    return yml_paths_all


def get_yml_paths(config: Json) -> Dict[str, Dict[str, Path]]:
    return get_workflow_paths(config, 'wic')


def get_py_paths(config: Json) -> Dict[str, Dict[str, Path]]:
    return get_workflow_paths(config, 'py')


def blindly_execute_python_workflows() -> None:
    """This function imports (read: blindly executes) all python files in 'search_paths_wic'
       The python files are assumed to have a top-level workflow() function
       which returns a sophios.api.pythonapi.Workflow object.
       The python files should NOT call the .run() method!
       (from any code path that is automatically executed on import)
    """
    # I hope u like Remote Code Execution vulnerabilities!
    # See https://en.wikipedia.org/wiki/Arithmetical_hierarchy
    from sophios.api import pythonapi  # pylint: disable=C0415:import-outside-toplevel
    # Since this is completely different test path we have to copy
    # default .txt files to default global_config.json
    config_file = Path().home()/'wic'/'global_config.json'
    global_config = io.get_default_config()
    pythonapi.global_config = get_tools_cwl(global_config)  # Use path fallback in the CI
    paths = get_py_paths(global_config)
    # Above we are assuming that config is default
    paths_tuples = [(path_str, path)
                    for namespace, paths_dict in paths.items()
                    for path_str, path in paths_dict.items()]
    any_import_errors = False
    for path_stem, path in paths_tuples:
        if 'mm-workflows' in str(path) or 'docs/tutorials/' in str(path):
            # Exclude paths that only contain 'regular' python files.
            continue
        # NOTE: Use anything (unique?) for the python_module_name.
        try:
            module = import_python_file(path_stem, path)
            # Let's require all python API files to define a function, say
            # def workflow() -> Workflow
            # so we can programmatically call it here:
            retval: pythonapi.Workflow = module.workflow()  # no arguments
            # which allows us to programmatically call Workflow methods:
            compiler_info = retval.compile()  # hopefully retval is actually a Workflow object!
            # But since this is python (i.e. not Haskell) that in no way eliminates
            # the above security considerations.

            # This lets us use path.parent to write a *.wic file in the
            # auto-discovery path, and thus reuse the existing wic CI
            retval.write_ast_to_disk(path.parent)

            # Programmatically blacklist subworkflows from running in config_ci.json
            # (Again, because subworkflows are missing inputs and cannot run.)
            config_ci = path.parent / 'config_ci.json'
            json_contents = {}
            if config_ci.exists():
                with open(config_ci, mode='r', encoding='utf-8') as r:
                    json_contents = json.load(r)
            run_blacklist: list[str] = json_contents.get('run_blacklist', [])
            # Use [1:] for proper subworkflows only
            subworkflows: list[pythonapi.Workflow] = retval.flatten_subworkflows()[1:]
            run_blacklist += [wf.process_name for wf in subworkflows]
            json_contents['run_blacklist'] = run_blacklist
            with open(config_ci, mode='w', encoding='utf-8') as f:
                json.dump(json_contents, f)

        except Exception as e:
            any_import_errors = True
            if sys.version_info >= (3, 10):
                traceback.print_exception(type(e), value=e, tb=None)
            else:
                traceback.print_exception(etype=type(e), value=e, tb=None)
    if any_import_errors:
        sys.exit(1)  # Make sure the CI fails
